import sys
from time import sleep, time
from threading import Thread
from functools import wraps
from contextlib import contextmanager

# Warning: signal is apparently unavailable on windows.
import signal

class Timeout(Exception): pass


@contextmanager
def timelimit(seconds, sig=signal.SIGALRM):
    """
    A decorator to limit a function to `timeout` seconds, raising `Timeout`.
    if it takes longer.

    >>> def meaningoflife():
    ...     sleep(.2)
    ...     return 42
    >>>
    >>> timelimit(.1)(meaningoflife)()
    Traceback (most recent call last):
        ...
    Timeout: Call took longer than 0.1 seconds.
    >>> timelimit(1)(meaningoflife)()
    42

    >>> with timelimit(.2):
    ...     sleep(1)
    Traceback (most recent call last):
        ...
    Timeout: Call took longer than 0.2 seconds.

    >>> with timelimit(.2):
    ...     sleep(.1)
    ...     print('finished')
    finished

    """
    if seconds is None:
        yield
        return

    def signal_handler(signum, frame):
        raise Timeout(f'Call took longer than {seconds} seconds.')

    signal.signal(sig, signal_handler)
    signal.setitimer(signal.ITIMER_REAL, seconds)
    yield
    signal.setitimer(signal.ITIMER_REAL, 0)   # disables alarm

timelimit.Timeout = Timeout

#_______________________________________________________________________________
#

def retry_apply(fn, args, kwargs=None, tries=2, pause=None, suppress=(Exception,),
                allow=(NameError, NotImplementedError)):
    """
    Attempt to call `fn` up to `tries` times with the `args` as arguments

    `suppress`: exceptions to be ignored up to the last retry-attempt.
    `allow`:    exceptions which will not be suppressed; this helps avoid
                stupid mistakes such as NameErrors and NotImplementedErrors.
    """
    if kwargs is None:
        kwargs = {}
    for i in range(tries):
        try:
            return fn(*args, **kwargs)
        except allow:          # raise these exceptions
            raise
        except suppress:
            if i == tries - 1:  # the last iteration
                raise
        if pause is not None: sleep(pause)


def retry(tries=2, pause=None, suppress=(Exception,), allow=(NameError, NotImplementedError)):
    def retry1(f):
        @wraps(f)
        def retry2(*args, **kw):
            return retry_apply(f, args, kw, tries=tries, pause=pause, suppress=suppress, allow=allow)
        return retry2
    return retry1


if __name__ == '__main__':
    from arsenal.assertions import assert_throws

    def test_retry():

        class NotCalledEnough(Exception): pass
        class TroublsomeFunction(object):
            "Function-like object which must be called >=4 times before succeeding."
            def __init__(self):
                self.tries = 0
            def __call__(self, *args):
                self.tries += 1
                if self.tries > 4:
                    return True
                else:
                    raise NotCalledEnough

        f = TroublsomeFunction()
        assert retry_apply(f, (1,2,3), tries=5)
        assert f.tries == 5

        with assert_throws(NotCalledEnough):
            f = TroublsomeFunction()
            print(retry_apply(f, (10,), tries=2))


        def create_trouble(tries_needs, attempts):
            calls = []
            @retry(tries=attempts, pause=0.0)
            def troublesome(a, kw=None):
                "I'm a troublesome function!"
                assert a == 'A' and kw == 'KW'
                calls.append(1)
                if len(calls) < tries_needs:
                    raise NotCalledEnough
                else:
                    return 'the secret'
            assert troublesome.__doc__ == "I'm a troublesome function!"
            assert troublesome.__name__ == 'troublesome'
            troublesome('A', kw='KW')

        create_trouble(4, 4)

        with assert_throws(NotCalledEnough):
            create_trouble(4, 2)


        @retry(tries=4, pause=0.0)
        def broken_function():
            raise NotImplementedError

        with assert_throws(NotImplementedError):
            broken_function()

        print('retry tests: pass')


    def test_timed():

        @timelimit(1.0)
        def sleepy_function(x): sleep(x)

        with assert_throws(Timeout):  # should timeout
            sleepy_function(3.0)

        sleepy_function(0.2)   # should succeed

        @timelimit(1)
        def raises_errors(): return 1/0
        with assert_throws(ZeroDivisionError):
            raises_errors()

        with timelimit(.2):
            sleep(.01)
            print('finished')

        tic = time()
        lim = .2
        with assert_throws(Timeout):
            with timelimit(lim):
                sleep(lim + 1)
        toc = time()
        took = toc - tic
        # make sure we wait at least `lim`.
        assert took > lim
        # make sure that there isn't too much overhead
        assert abs(lim - took) < 0.001, abs(lim - took)
        print('decorator tests: pass')

    test_retry()
    test_timed()

    import doctest
    doctest.testmod(verbose=True)

    print('passed tests..')



#!/usr/bin/env python
"""
Find most informative features ranked by information gain (i.e. the ID3 hueristic for decision trees).

Input: a tab-delimited file where each line starts with a label followed by features.
Output: a sorted list of the most informative features.
"""

import sys
from numpy import dot, log, zeros, fromiter, int32
from collections import defaultdict


# Had to dump Alphabet directly into this file because it was taking too long to
# import arsenal stuff... thanks to scipy stuff being brought in... sigh.
import os
from numpy.random import randint


class Alphabet(object):
    """
    Bijective mapping from strings to integers.

    >>> a = Alphabet()
    >>> [a[x] for x in 'abcd']
    [0, 1, 2, 3]
    >>> list(map(a.lookup, range(4)))
    ['a', 'b', 'c', 'd']

    >>> a.stop_growth()
    >>> a['e']

    >>> a.freeze()
    >>> a.add('z')
    Traceback (most recent call last):
      ...
    ValueError: Alphabet is frozen. Key "z" not found.

    >>> print(a.plaintext())
    a
    b
    c
    d
    """

    def __init__(self, random_int=None):
        self._mapping = {}   # str -> int
        self._flip = {}      # int -> str; timv: consider using array or list
        self._i = 0
        self._frozen = False
        self._growing = True
        self._random_int = random_int   # if non-zero, will randomly assign
                                        # integers (between 0 and randon_int) as
                                        # index (possibly with collisions)

    def __repr__(self):
        return 'Alphabet(size=%s,frozen=%s)' % (len(self), self._frozen)

    def freeze(self):
        self._frozen = True

    def stop_growth(self):
        self._growing = False

    @classmethod
    def from_iterable(cls, s):
        "Assumes keys are strings."
        inst = cls()
        for x in s:
            inst.add(x)
#        inst.freeze()
        return inst

    def keys(self):
        return iter(self._mapping.keys())

    def items(self):
        return iter(self._mapping.items())

    def imap(self, seq, emit_none=False):
        """
        Apply alphabet to sequence while filtering. By default, `None` is not
        emitted, so the Note that the output sequence may have fewer items.
        """
        if emit_none:
            for s in seq:
                yield self[s]
        else:
            for s in seq:
                x = self[s]
                if x is not None:
                    yield x

    def map(self, seq, *args, **kwargs):
        return list(self.imap(seq, *args, **kwargs))

    def add_many(self, x):
        for k in x:
            self.add(k)

    def lookup(self, i):
        if i is None:
            return None
        #assert isinstance(i, int)
        return self._flip[i]

    def lookup_many(self, x):
        for k in x:
            yield self.lookup(k)

    def __contains__(self, k):
        #assert isinstance(k, basestring)
        return k in self._mapping

    def __getitem__(self, k):
        try:
            return self._mapping[k]
        except KeyError:
            #if not isinstance(k, basestring):
            #    raise ValueError("Invalid key (%s): only strings allowed." % (k,))
            if self._frozen:
                raise ValueError('Alphabet is frozen. Key "%s" not found.' % (k,))
            if not self._growing:
                return None
            if self._random_int:
                x = self._mapping[k] = randint(0, self._random_int)
            else:
                x = self._mapping[k] = self._i
                self._i += 1
            self._flip[x] = k
            return x

    add = __getitem__

    def __setitem__(self, k, v):
        assert k not in self._mapping
        assert isinstance(v, int)
        self._mapping[k] = v
        self._flip[v] = k

    def __iter__(self):
        for i in range(len(self)):
            yield self._flip[i]

    def enum(self):
        for i in range(len(self)):
            yield (i, self._flip[i])

    def __len__(self):
        return len(self._mapping)

    def plaintext(self):
        "assumes keys are strings"
        return '\n'.join(self)


def normalize(p):
    return p / p.sum()


def lidstone(p, delta):
    """
    Lidstone smoothing is a generalization of Laplace smoothing.
    """
    return normalize(p + delta)


def kl_divergence(p, q):
    """ Compute KL divergence of two vectors, K(p || q).
    NOTE: If any value in q is 0.0 then the KL-divergence is infinite.
    """
    return dot(p, log(p) - log(q))


def read_tab_file(f):
    """
    Load really simple tab-separated file format for labeled data. Label is the
    first column, features are the remaining columns.
    """
    for line in f:
        line = line.strip()
        line = line.split('\t')
        if len(line) <= 1:
            continue
        yield line[0], line[1:]


def integerize(data):
    """
    Integerize dataset
    returns a triple (label alphabet, feature alphabet, integerized dataset)
    """
    F = Alphabet()
    L = Alphabet()
    I = [(L[label], fromiter(F.map(features), dtype=int32)) for label, features in data]
    return (L, F, I)



def kl_filter(data,
              verbose=True,
              progress=False,
              out=sys.stdout,
              feature_label_cuttoff=0,
              feature_count_cuttoff=0,
              do_label_count=False):
    """
    data = (label, [features ...])

    KL is a synonym for Information Gain

    KL( p(label) || p(label|feature) )
    """
    (L, F, data) = integerize(data)

    if do_label_count:
        label_count = defaultdict(int)
        for label, _ in data:
            label_count[label] += 1
        label_count = list(label_count.items())
        label_count.sort(key=lambda x: -x[1])  # sort by count
        print('label count')
        for k,v in label_count:
            print('%20s => %s' % (k, v))
        sys.exit(0)

    K = len(L)
    M = len(F)

    if progress:
        from arsenal.iterextras import iterview
    else:
        iterview = lambda x, *a, **kw: x

    if progress:
        print('\nTally', file=sys.stderr)

    # label-feature tally (note: we ignore dulicate features)
    counts = zeros((K,M))
    for y, fv in iterview(data):
        counts[y, fv] += 1

    feature_counts = counts.sum(axis=0)

    if feature_count_cuttoff > 0:
        cut = feature_counts < feature_count_cuttoff

        #if verbose:
        print('%s of %s below cutoff of %s' \
            % (cut.sum(), len(feature_counts), feature_count_cuttoff), file=sys.stderr)

        if progress:
            print('%s / %s (%.2f%%) features below cuttoff' % \
                (cut.sum(), M, cut.sum()*100.0/M), file=sys.stderr)

        # zero-out features below cuttoff
        counts[:, cut] = 0

    if feature_label_cuttoff:
        cut = counts < feature_label_cuttoff

        if progress:
            print('%s / %s (%.2f%%) feature-label pairs below cuttoff' % \
                (cut.sum(), K*M, cut.sum()*100.0/(K*M)), file=sys.stderr)

        # zero-out features below cuffoff
        counts[cut] = 0

    label_prior = lidstone(counts.sum(axis=1), 0.001)  # avoids divide-by-zero

    # compute KL
    if progress:
        print('\nKL', file=sys.stderr)

    KL = zeros(M)
    for f in iterview(range(M)):
        label_given_f = lidstone(counts[:,f], 0.001)   # avoids divide-by-zero
        KL[f] = -kl_divergence(label_prior, label_given_f)

    # print KL-feature, most-informative first
    for i in KL.argsort():

        z = counts[:,i].sum()

        if z == 0:
            continue

        p = counts[:,i] * 1.0 / z

        l = [(v, k) for k,v in zip(L, p) if v > 0]
        l.sort()

        z = (-KL[i], F.lookup(i), l)

        if verbose:
            print('%8.6f\t%s\t%s' % (-KL[i], int(counts[:,i].sum()), F.lookup(i)), '\t\033[32m', ' '.join('%s(%.4f)' % (k,v) for v, k in l), '\033[0m', file=out)

        yield z


if __name__ == '__main__':
    from argparse import ArgumentParser
    p = ArgumentParser()
    p.add_argument('input', type=open)
    p.add_argument('--feature-cuttoff', type=int, default=0)
    p.add_argument('--feature-label-cuttoff', type=int, default=0)
    p.add_argument('--progress', action='store_true')
    args = p.parse_args()

    #list(kl_filter(read_tab_file(file(sys.argv[1]) if len(sys.argv) == 2 else sys.stdin)))

    for _ in kl_filter(read_tab_file(args.input),
                       feature_count_cuttoff=args.feature_cuttoff,
                       feature_label_cuttoff=args.feature_label_cuttoff,
                       progress=args.progress):
        pass



from itertools import cycle, islice


def _fair_product(A, B):
    """
    Cartesian product of two possibly infinite iterables
    Based closely on https://github.com/sympy/sympy/issues/17157
    """

    A = iter(A)
    B = iter(B)

    a = []
    b = []

    sentinel = object()
    def append(it, elems):
        e = next(it, sentinel)
        if e is not sentinel:
            elems.append(e)

    n = 0
    append(A, a)
    append(B, b)

    while n <= len(a) + len(b):
        for m in range(n-len(a)+1, len(b)):
            yield (a[n-m], b[m])
        n += 1
        append(A, a)
        append(B, b)


def fair_product(*iterables):
    """
    Returns Cartesian product of iterables. Yields every element
    eventually.

    Based closely on https://github.com/sympy/sympy/issues/17157
    """
    if len(iterables) == 0:
        yield ()
        return
    elif len(iterables) == 1:
        for e in iterables[0]:
            yield (e,)
    elif len(iterables) == 2:
        for e12 in _fair_product(*iterables):
            yield e12
    else:
        first, *others = iterables
        for ef, eo in fair_product(first, fair_product(*others)):
            yield (ef,) + eo



def merge_roundrobin(*iterables):
    """
    Merge iterators with a round-robin scheduler.
    Original implementation by George Sakkis.

    >>> list(merge_roundrobin('ABC', 'D', 'EF'))
    ['A', 'D', 'E', 'B', 'F', 'C']

    >>> from itertools import count
    >>> from arsenal.iterextras.util import take
    >>> list(take(10, merge_roundrobin('ABC', count(), 'E')))
    ['A', 0, 'E', 'B', 1, 'C', 2, 3, 4, 5]

    """
    pending = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while pending:
        try:
            for _next in nexts:
                yield _next()
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))

fair_union = merge_roundrobin


def test_fair_product():
    from itertools import count

    def show_cross_product(xys, steps, N):
        grid = [['  .' for j in range(N)] for i in range(N)]
        for i, (x, y) in enumerate(xys):
            if i >= steps: break
            try:
                grid[x][y] = f'{i:3d}'
            except IndexError:
                pass
        print('[%s]' % ( ']\n['.join(' '.join(map(str,row)) for row in grid)))


    N = 10
    steps = N*(N+1)//2

    from arsenal import colors
    print()
    print(colors.light.yellow % 'fair_product: inf-inf')
    show_cross_product(fair_product(count(), count()),
                       steps = steps, N = N)

    print()
    print(colors.light.yellow % 'fair_product: inf-fin')
    show_cross_product(fair_product(count(), range(4)),
                       steps = steps, N = N)

    print()
    print(colors.light.yellow % 'fair_product: fin-inf')
    show_cross_product(fair_product(range(4), count()),
                       steps = steps, N = N)

    print()
    print(colors.light.yellow % 'fair_product: fin-fin')
    show_cross_product(fair_product(range(3), range(5)),
                       steps = steps, N = N)


if __name__ == '__main__':
    test_fair_product()

    import doctest
    doctest.testmod()



import atexit
import shelve
import pickle as pickle

from functools import partial


# TODO:
#  * add option to pass a reference to another cache
class memoize(object):
    """ cache a function's return value to avoid recalulation """
    def __init__(self, func):
        self.func = func
        self.cache = {}
        try:
            self.__name__ = func.__name__
            self.__doc__ = func.__doc__
        except AttributeError:
            pass

    def __get__(self, obj, objtype=None):
        "define `__get__` in case this function is a method."
        #print('  method get', obj, objtype)
        if obj is None: return self.func
        return partial(self, obj)

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            try:
                self.cache[args] = value
            except TypeError:
                # uncachable -- for instance, passing a list as an argument.
                raise TypeError('uncachable arguments %r passed to memoized function.' % (args,))
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            raise TypeError('uncachable arguments %r passed to memoized function.' % (args,))

    def __repr__(self):
        return '<memoize(%r)>' % self.func


class ShelfBasedCache(object):
    """ cache a function's return value to avoid recalulation and save cache in a shelve. """
    def __init__(self, func, key, None_is_bad=False):
        self.func = func
        self.filename = '{self.func.__name__}.shelf~'.format(self=self)
        self.cache = shelve.open(self.filename, flag='c') #, writeback=True)
        self.key = key
        self.None_is_bad = None_is_bad
        self.__name__ = 'ShelfBasedCache(%s)' % func.__name__
    def __call__(self, *args):
        p_args = self.key(args)
        value = None
        recompute = True
        if self.cache.has_key(p_args):
            recompute = False
            value = self.cache[p_args]
            if value is None and self.None_is_bad:
                recompute = True
        if recompute:
            self.cache[p_args] = value = self.func(*args)
            self.cache.sync()
        return value

def persistent_cache(key=lambda x: x, None_is_bad=False):
    def wrap(f):
        return ShelfBasedCache(f, key, None_is_bad=None_is_bad)
    return wrap



## TODO: automatically make a back-up of any previous pickles just in case the
##   save fails.  (Saving at-exit can be pretty flaky.)
class memoize_persistent(object):
    """
    cache a function's return value to avoid recalulation and save the
    cache (via pickle) at system exit so that it persists.

    WARNING: retrieves cache for functions which might not be equivalent
             if a revision is made to the code which is used to compute it.
    """
    def __init__(self, func, filename=None):
        self.func = func
        self.filename = filename or '{self.func.__name__}.cache.pkl~'.format(self=self)
        self.dirty = False
        self.key = 0
        self.cache = {}
        self.loaded = False
        atexit.register(self.save)

    def save(self):
        if self.cache and self.dirty:
            with open(self.filename, 'wb') as f:
                pickle.dump((self.cache, self.key), f)
            print('[ATEXIT] saved persistent cache for {self.func.__name__} to file "{self.filename}"'.format(self=self))
        else:
            print("[ATEXIT] found nothing to save in {self.func.__name__}'s cache.".format(self=self))

    def load(self):
        self.loaded = True
        loaded_key = None
        try:
            with open(self.filename, 'rb') as f:
                (cache, loaded_key) = pickle.load(f)
        except IOError:
            pass
        finally:
            if self.key == loaded_key:
                self.cache = cache
                #print 'loaded cache for {self.func.__name__}'.format(self=self)
            else:
                self.cache = {}
                #print 'failed to load cache for {self.func.__name__}'.format(self=self)

    def __call__(self, *args):
        # wait until you call the function to un-pickle
        if not self.loaded:
            self.load()
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            try:
                self.cache[args] = value
            except TypeError:
                # uncachable -- for instance, passing a list as an argument.
                raise TypeError('uncachable arguments %r passed to memoized function.' % (args,))
            else:
                self.dirty = True
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            raise TypeError('uncachable arguments %r passed to memoized function.' % (args,))

    def get_cached(self, *args):
        """ If result is cached return it, otherwise return `None`. """
        # wait until you call the function to un-pickle
        if not self.loaded:
            self.load()
        if args in self.cache:
            return self.cache[args]
        else:
            return None


def test_memoize():

    @memoize
    def g(x):
        return x**2

    class foo:
        def __init__(self, a):
            self.a = a
        @memoize
        def goo(self, x):
            return self.a * x
        def __repr__(self):
            return f'foo({self.a})'

    a = foo(2)
    b = foo(3)

    print('created')
    a_goo = a.goo
    print(a.goo)
    print('calling')
    assert a_goo(5) == 2*5
    assert b.goo(5) == 3*5
    print('ok')

    assert a_goo(5) == 2*5

    print('xxx')
    print(foo.goo)   # triggers method.get
    assert foo.goo(a, 4) == 2*4
    print('----')

    print(a.__class__.__dict__['goo'])        # gives the memoize instance
    print(a.__class__.__dict__['goo'].cache)  # it's hashing (obj, args)

    print('----')
    assert g(4) == 4**2    # no __get__ call here.
    print(g)
    print(g.cache)
    print('ok')


if __name__ == '__main__':
    test_memoize()



cdef class Vector:

    cdef public:
        int cap
        int end
        double[:] val

    cpdef int push(self, double x)
    cpdef object pop(self)
    cdef void grow(self)
    cdef void ensure_size(self, int i)
    cdef double get(self, int i)
    cdef void set(self, int i, double v)


cdef class MaxHeap:

    cdef public:
        Vector val

#    def pop(self)
#    def peek(self)
#    def push(self, v)
    cdef void swap(self, int i, int j)
    cdef int up(self, int i)
    cdef int down(self, int i)
    cdef int _update(self, int i, double old, double new)
    cdef void _remove(self, int i)


cdef class LocatorMaxHeap(MaxHeap):
    """
    Dynamic heap. Maintains max of a map, via incrementally maintained partial
    aggregation tree. Also known a priority queue with 'locators'.

    This data structure efficiently maintains maximum of the priorities of a set
    of keys. Priorites may increase or decrease. (Many max-heap implementations
    only allow increasing priority.)

    """

    cdef public:
        dict key
        dict loc

#    def pop(self)
#    def popitem(self)
#    def peek(self)
    cdef void _remove(self, int i)
    cdef _setitem(self, object k, double v)
#    def __delitem__(self, k)
#    def __contains__(self, k)
#    def __getitem__(self, k)
#    def __setitem__(self, k, v)
    cdef void swap(self, int i, int j)
#    def check(self)
#    cdef _update(self, int i, double old, double new)



# -*- coding: utf-8 -*-
import re, os, sys
from glob import glob
from subprocess import Popen, PIPE


def ansi2html(x):
    p = Popen('ansi2html.sh', stdout=PIPE, stdin=PIPE, stderr=PIPE)
    p.stdin.write(str(x).encode('utf-8'))
    return p.communicate()[0].decode('utf-8')


def link(url, text=None):
    if text is None:
        text = url
    return f"\033]8;;{url}\a{text}\033]8;;\a"


def overline(xs):
    return ''.join(f'{x}\u0305' for x in xs)


partial = '∂'

visible_space = '␣'

class bb:
    A = '𝔸'
    B = '𝔹'
    C = 'ℂ'
    D = '𝔻'
    E = '𝔼'
    F = '𝔽'
    G = '𝔾'
    H = 'ℍ'
    I = '𝕀'
    J = '𝕁'
    K = '𝕂'
    L = '𝕃'
    M = '𝕄'
    N = 'ℕ'
    O = '𝕆'
    P = 'ℙ'
    Q = 'ℚ'
    R = 'ℝ'
    S = '𝕊'
    T = '𝕋'
    U = '𝕌'
    V = '𝕍'
    W = '𝕎'
    X = '𝕏'
    Y = '𝕐'
    Z = 'ℤ'


# https://stackoverflow.com/questions/8651361/how-do-you-print-superscript-in-python
superscript_map = {
    "0": "⁰", "1": "¹", "2": "²", "3": "³", "4": "⁴", "5": "⁵", "6": "⁶",
    "7": "⁷", "8": "⁸", "9": "⁹",
    0: "⁰", 1: "¹", 2: "²", 3: "³", 4: "⁴", 5: "⁵", 6: "⁶",
    7: "⁷", 8: "⁸", 9: "⁹",
    "a": "ᵃ", "b": "ᵇ", "c": "ᶜ", "d": "ᵈ",
    "e": "ᵉ", "f": "ᶠ", "g": "ᵍ", "h": "ʰ", "i": "ᶦ", "j": "ʲ", "k": "ᵏ",
    "l": "ˡ", "m": "ᵐ", "n": "ⁿ", "o": "ᵒ", "p": "ᵖ", "q": "۹", "r": "ʳ",
    "s": "ˢ", "t": "ᵗ", "u": "ᵘ", "v": "ᵛ", "w": "ʷ", "x": "ˣ", "y": "ʸ",
    "z": "ᶻ", "A": "ᴬ", "B": "ᴮ", "C": "ᶜ", "D": "ᴰ", "E": "ᴱ", "F": "ᶠ",
    "G": "ᴳ", "H": "ᴴ", "I": "ᴵ", "J": "ᴶ", "K": "ᴷ", "L": "ᴸ", "M": "ᴹ",
    "N": "ᴺ", "O": "ᴼ", "P": "ᴾ", "Q": "Q", "R": "ᴿ", "S": "ˢ", "T": "ᵀ",
    "U": "ᵁ", "V": "ⱽ", "W": "ᵂ", "X": "ˣ", "Y": "ʸ", "Z": "ᶻ", "+": "⁺",
    "-": "⁻", "=": "⁼", "(": "⁽", ")": "⁾",
}

subscript_map = {
    "0": "₀", "1": "₁", "2": "₂", "3": "₃", "4": "₄", "5": "₅", "6": "₆",
    "7": "₇", "8": "₈", "9": "₉",
    0: "₀", 1: "₁", 2: "₂", 3: "₃", 4: "₄", 5: "₅", 6: "₆",
    7: "₇", 8: "₈", 9: "₉",
    "a": "ₐ", "b": "♭", "c": "꜀", "d": "ᑯ",
    "e": "ₑ", "f": "բ", "g": "₉", "h": "ₕ", "i": "ᵢ", "j": "ⱼ", "k": "ₖ",
    "l": "ₗ", "m": "ₘ", "n": "ₙ", "o": "ₒ", "p": "ₚ", "q": "૧", "r": "ᵣ",
    "s": "ₛ", "t": "ₜ", "u": "ᵤ", "v": "ᵥ", "w": "w", "x": "ₓ", "y": "ᵧ",
    "z": "₂", "A": "ₐ", "B": "₈", "C": "C", "D": "D", "E": "ₑ", "F": "բ",
    "G": "G", "H": "ₕ", "I": "ᵢ", "J": "ⱼ", "K": "ₖ", "L": "ₗ", "M": "ₘ",
    "N": "ₙ", "O": "ₒ", "P": "ₚ", "Q": "Q", "R": "ᵣ", "S": "ₛ", "T": "ₜ",
    "U": "ᵤ", "V": "ᵥ", "W": "w", "X": "ₓ", "Y": "ᵧ", "Z": "Z", "+": "₊",
    "-": "₋", "=": "₌", "(": "₍", ")": "₎",
}

class arrow:
    u = up = '↑'
    ur = '↗'
    r = right = '→'
    dr = '↘'
    dl = '↙'
    d = down = '↓'
    l = left =  '←'
    ul = '↖'

    class long:
        right = r = '⟶'
        leftright = lr = '⟷'
        left = l = '⟵'

    class Long:
        leftright = lr = '⟺'
        left = l = '⟸'
        right = r = '⟹'

    class squiggle:
        right = r = '⟿'


arrows = arrow.__dict__

def superscript(x):
    return "".join(superscript_map[i] for i in str(x))

def subscript(x):
    return "".join(subscript_map[i] for i in str(x))


def complete_filenames(text, line, begidx, endidx):
    "Util for filename completion."

    before_arg = line.rfind(" ", 0, begidx)
    if before_arg == -1:
        return # arg not found

    fixed = line[before_arg+1:begidx]  # fixed portion of the arg

    completions = []
    for p in glob(line[before_arg+1:endidx] + '*'):
        p = p + (os.sep if p and os.path.isdir(p) and p[-1] != os.sep else '')
        completions.append(p.replace(fixed, "", 1))
    return completions


def ansi(color=None, light=None, bg=3):
    return '\x1b[%s;%s%sm' % (light, bg, color)

_reset = '\x1b[0m'


def colorstring(s, c):
    return c + s + _reset



black, red, green, yellow, blue, magenta, cyan, white = \
    [colorstring('%s', ansi(c, 0)) for c in range(8)]

class light:
    black, red, green, yellow, blue, magenta, cyan, white = \
        [colorstring('%s', ansi(c, 1)) for c in range(8)]

class dark:
    black, red, green, yellow, blue, magenta, cyan, white = \
        [colorstring('%s', ansi(c, 2)) for c in range(8)]

class bg:
    black, red, green, yellow, blue, magenta, cyan, white = \
        [colorstring('%s', ansi(c, 0, bg=4)) for c in range(8)]

    def rgb(r,g,b):
        return f"\x1b[48;2;{r};{g};{b}m%s\x1b[0m"

    orange = rgb(255, 165, 0)
    rgb = staticmethod(rgb)

def rgb(r,g,b):
    return f"\x1b[38;2;{r};{g};{b}m%s\x1b[0m"

orange = rgb(255, 165, 0)

purple = '\x1b[38;5;91m' + '%s' + _reset


normal = '\x1b[0m%s\x1b[0m'
bold = '\x1b[1m%s\x1b[0m'
italic = "\x1b[3m%s\x1b[0m"
underline = "\x1b[4m%s\x1b[0m"
strike = "\x1b[9m%s\x1b[0m"
#overline = lambda x: (u''.join(unicode(c) + u'\u0305' for c in unicode(x))).encode('utf-8')

leftarrow = '←'
rightarrow = '→'
reset = _reset

ansi2html = ansi2html


def nocolor(x):
    "Remove ANSI control codes from the given string `x`."
    return re.sub('\033\[[0-9;]*m', "", x)


# https://www.w3.org/TR/xml-entity-names/025.html
def branch(xs):
    ys = []
    for i, x in enumerate(xs):
        first = i == 0
        last = i == len(xs)-1
        if last and first:
            h = '└─'
        elif last:
            h = '└─'
        elif first:
            h = '┌─' #'├─'
        else:
            h = '├─'
        if not x:
            continue
        ys.append(h + x[0])
        indent = '│ ' if not last else '  '
        for a in x[1:]:
            ys.append(indent + a)
    return ys


def line(n): return '─'*(n)

def thick_line(n): return ('━'*80)

def mark_(x): return '✔' if x else '✘'

def faster(x, y, color=True):
    "is x faster than y"
    msg = f'{x / y:.2f}x faster' if x > y else f'{y / x:.2f}x slower'
    msg = f'{msg} ({x:g} s vs {y:g} s)'
    if color: msg = (green if x > y else yellow) % msg
    return msg

from textwrap import dedent

def indent(indentation, msg):
    if isinstance(msg, str): msg = msg.split('\n')
    return '\n'.join(indentation + line for line in msg)


def reindent(indentation, msg):
    return indent(indentation, dedent(msg))

#def padr(w):
#    "get format to pad right elements"
#    return '%%%ss' % w
#pad = padr
#def padl(w):
#    "get format to pad left elements"
#    return '%%-%ss' % w
#def getwidth(a):
#    "Find maximum width of the string representations of the elements of ``a``."
#    return max(len(str(z)) for z in a)


def check(x, t='pass', f='fail'):
    return green % t if x else red % f


#def color01(x, fmt='%.10f', min_color=235, max_color=255):
#    "Colorize numbers in [0,1] based on value; darker means smaller value."
#    import colored
#    if not (0 <= x <= 1 + 1e-10):
#        return colors.red % fmt % x
#    width = max_color - min_color
#    color = min_color + int(round(x*width))
#    return '%s%s%s' % (colored.fg(color), (fmt % x), colored.attr('reset'))


def console_width(minimum=None, default=80):
    "Return width of available window area."
    from tqdm.utils import _screen_shape_wrapper
    x = _screen_shape_wrapper()(sys.stdout)[0]
    return max(minimum or 0, x or default)


def marquee(msg=''):
    return ('{0:*^%s}' % console_width()).format(msg)


import re
def render(y, **kwargs):
    """
    Render colorful string using 'reset' to mean 'pop the color stack' rather than
    go directly 'normal' color.
    """
    return rendering(y, **kwargs).value


class rendering:
    def __init__(self, y, debug=False):
        y = str(y)
        xs = re.split('(\x1b\[[0-9;]+m)', y)  # tokenize.

        if debug:
            msg = lambda *args: print(f'[render]{"    "*len(s)}', *args)

        s = [_reset]      # stack
        b = []            # buffer
        prefix = '\x1b['  # control code prefix
        c = _reset        # current color
        for x in xs:
            if debug: msg('current token', repr(x))
            if x.startswith(prefix):
                if x == _reset:
                    if debug: msg('   ^ pop')
                    c = s.pop() if len(s) else _reset
                else:
                    if debug: msg('   ^ control code', x + repr(x) + _reset)
                    s.append(c)
                    c = x
                continue
            else:
                if debug: msg('   ^ color', c + repr(x) + _reset)
                b.append(c)
                b.append(x + _reset)
        b.append(_reset)   # always end on reset.
        self.xs = xs
        self.value = ''.join(b)

    def __len__(self):
        return sum(len(x) for x in self.xs if not x.startswith('\x1b['))  # control code prefix

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    def __format__(self, spec):
        [(pad, fmt)] = re.findall('^(-?\d*)(s)$', spec)
        pad = int(pad)
        if pad > 0:
            return self.value + ' '*max(0, pad - len(self))
        else:
            return ' '*max(0, abs(pad) - len(self)) + self.value


ok   = green     % 'ok'
warn = yellow    % 'warn'
fail = light.red % 'fail'
bad  = light.red % 'bad'
error  = light.red % 'error'


thumbs_up = '👍'
thumbs_down = '👎'
poop = poo = turd = '💩'
timeout = '⌛'

trophy = '🏆'

lightbulb = '💡'
fire = '🔥'
lightning = '⚡'

#lightning = '⚡'
#lightning = '↯'
#lightning = 'ϟ'

check = green % '✔' #'✓'
xmark = dark.red % '✘' # ☒ - Unicode Character Table
def mark(x): return check if x else xmark

def percent(num, den):
    if den == 0:
        return f'{100:.1f}% (0/0)'
    return f'{num*100/den:.1f}% ({num}/{den})'
percentage = percent


def tests():
    for c in 'black, red, green, yellow, blue, magenta, cyan, white'.split(', '):
        print('%18s %24s %23s %21s' % (globals()[c] % c,
                                       getattr(light, c) % f'light.{c}',
                                       getattr(dark, c) % f'dark.{c}',
                                       getattr(bg, c) % f'bg.{c}'))

    print(underline % 'underline')
    print(italic % 'italic')
    print(strike % 'strike')

    #import numpy as np
    #for x in np.linspace(0, 1, 15):
    #    print(color01(x, fmt='%.2f'), end=' ')
    #print()

    w = console_width()
    print('Console width:', w, sep=' ')
    print('='*w)
    print(marquee(' marquee '))


    print()
    print('Stack-based rendering')
    print('=====================')
    g = green % 'green'
    b = blue % f'blue {g} blue'
    r = red % f'red {b} red {b} red'
    x = normal % f'normal {r} normal'
    print(render(x))

    # SDD: make sure we reset after dark
    print(render(
        (light.blue % 'light %s light'
         % dark.blue % 'dark %s dark'
         % blue % 'regular'),
        #debug = True,
    ))

    print(thumbs_up, thumbs_down)


if __name__ == '__main__':
    tests()



class MultiMap:
    """
    Multi-dimensional map data structure.
    """

    def __init__(self, vals=None):
        self.vals = {} if vals is None else vals
        self.index = {}

    def __iter__(self):
        return iter(self.vals)

    def __setitem__(self, item, val):
        if not isinstance(item, (list, tuple)): item = (item,)
        assert not any(isinstance(y, slice) for y in item), 'setting range values not supported'
        if item not in self.vals:
            # We only need to update indexes when we get a new item because
            # indexes only track the support.  To update, we loop through all
            # active indexes (i.e., query patterns).
            for m,ix in self.index.items():
                k = tuple(item[i] for i in m)
                if k not in ix: ix[k] = set()
                ix[k].add(item)
        self.vals[item] = val

    def __getitem__(self, query):
        if not isinstance(query, (list, tuple)): query = (query,)
        for y in query:
            if isinstance(y, slice):
                assert y == slice(None, None, None), 'only simple slices allowed.'
        m = tuple(i for i, y in enumerate(query) if not isinstance(y, slice))
        k = tuple(y for i, y in enumerate(query) if not isinstance(y, slice))
        if m not in self.index: self._index(m)
        return MultiMap({x: self.vals[x] for x in self.index[m].get(k, set())
                         if len(x) == len(query)})

    def _index(self, m):
        self.index[m] = ix = {}
        for x in self.vals:
            k = tuple(x[i] for i in m)
            if k not in ix: ix[k] = set()
            ix[k].add(x)
        return ix

    def __repr__(self):
        return f'MultiMap({self.vals})'

    def __eq__(self, other):
        return self.vals == other.vals

    def __str__(self):
        # Sort if types allow it, otherwise fall back to the order in vals dictionary
        try:
            vs = sorted(self.vals)
        except TypeError:
            vs = self.vals
        return 'MultiMap {\n%s\n}' % '\n'.join(f'  {x}: {self.vals[x]},' for x in vs)


def tests_basics():
    m = MultiMap()
    m["a","b","c"] = 10
    m["a","b'","c"] = 12

    m["a","b","d"] = 13
    m["a", 14,frozenset(),None,()] = 13   # ragged dimensions allowed (YMMV), dims can have mixed types

    from arsenal.assertions import assert_throws
    with assert_throws(AssertionError):
        m["a",:,"d"] = 13
    with assert_throws(AssertionError):
        print(m["a",:10,"d"])

    assert m["a",:,"c"] == MultiMap({('a', 'b', 'c'): 10, ('a', "b'", 'c'): 12})
    assert m[:,:,"d"] == MultiMap({('a', 'b', 'd'): 13})

    print(m)

    # Note: This does not grab prefixes,
    # print(m["a",:,:])

    # Annoying corner cases for when we don't pass a tuple to get/set items
    m = MultiMap()
    m[1] = 2
    assert m[1,] == m[1] == m
    m[1,] = 2
    assert m[1,] == m[1] == m

    assert list(m) == [(1,)]


if __name__ == '__main__':
    tests_basics()



# cython: language_level=3, boundscheck=False, infer_types=True, nonecheck=False
# cython: overflowcheck=False, initializedcheck=False, wraparound=False, cdivision=True

"""
Heap data structures with optional
 - Locators
 - Top-k (Bounded heap)

"""
import numpy as np

Vt = np.double
cdef double NaN = np.nan

# TODO: Use the C++ standard library's implementation of a vector of doubles.
cdef class Vector:

#    cdef public:
#        int cap
#        int end
#        double[:] val

    def __init__(self, cap):
        self.cap = cap
        self.val = np.zeros(self.cap, dtype=Vt)
        self.end = 0

    cpdef int push(self, double x):
        i = self.end
        self.ensure_size(i)
        self.val[i] = x
        self.end += 1
        return i

    cpdef object pop(self):
        "pop from the end"
        assert 0 < self.end
        self.end -= 1
        v = self.val[self.end]
        self.val[self.end] = NaN
        return v

    cdef void grow(self):
        self.cap *= 2
        new = np.empty(self.cap, dtype=Vt)
        new[:self.end] = self.val[:self.end]
        self.val = new

    cdef void ensure_size(self, int i):
        "grow in needed"
        if self.val.shape[0] < i + 1: self.grow()

    def __getitem__(self, int i):
        assert i < self.end
        return self.get(i)

    cdef double get(self, int i):
        return self.val[i]

    def __setitem__(self, int i, double v):
        assert i < self.end
        self.set(i, v)

    cdef void set(self, int i, double v):
        self.val[i] = v

    def __len__(self):
        return self.end

    def __repr__(self):
        return repr(self.val[:self.end])


cdef class MaxHeap:

#    cdef public:
#        Vector val

    def __init__(self, cap=2**8):
        self.val = Vector(cap)
        self.val.push(np.nan)

    def __len__(self):
        return len(self.val) - 1   # subtract one for dummy root element

    def pop(self):
        v = self.peek()
        self._remove(1)
        return v

    def peek(self):
        return self.val.val[1]

    def push(self, v):
        # put new element last and bubble up
        return self.up(self.val.push(v))

    cdef void swap(self, int i, int j):
        assert i < self.val.end
        assert j < self.val.end
        self.val.val[i], self.val.val[j] = self.val.val[j], self.val.val[i]

    cdef int up(self, int i):
        while 1 < i:
            p = i // 2
            if self.val.val[p] < self.val.val[i]:
                self.swap(i, p)
                i = p
            else:
                break
        return i

    cdef int down(self, int i):
        n = self.val.end
        while 2*i < n:
            a = 2 * i
            b = 2 * i + 1
            c = i
            if self.val.val[c] < self.val.val[a]:
                c = a
            if b < n and self.val.val[c] < self.val.val[b]:
                c = b
            if c == i:
                break
            self.swap(i, c)
            i = c
        return i

    cdef int _update(self, int i, double old, double new):
        assert i < self.val.end
        if old == new: return i   # value unchanged
        self.val.val[i] = new         # perform change
        if old < new:             # increased
            return self.up(i)
        else:                     # decreased
            return self.down(i)

    cdef void _remove(self, int i):
        # update the locator stuff for last -> i
        last = self.val.end - 1
        self.swap(i, last)
        old = self.val.pop()
        # special handling for when the heap has size one.
        if i == last: return
        self._update(i, old, self.val.val[i])

    def check(self):
        # heap property
        for i in range(2, self.val.end):
            assert self.val[i] <= self.val[i // 2], (self.val[i // 2], self.val[i])   # child <= parent


cdef class LocatorMaxHeap(MaxHeap):
    """
    Dynamic heap. Maintains max of a map, via incrementally maintained partial
    aggregation tree. Also known a priority queue with 'locators'.

    This data structure efficiently maintains maximum of the priorities of a set
    of keys. Priorites may increase or decrease. (Many max-heap implementations
    only allow increasing priority.)

    """

#    cdef public:
#        dict key
#        dict loc

    def __init__(self, **kw):
        super().__init__(**kw)
        self.key = {}   # map from index `i` to `key`
        self.loc = {}   # map from `key` to index in `val`

    def __repr__(self):
        return repr({k: self[k] for k in self.loc})

    def pop(self):
        k,v = self.peek()
        super().pop()
        return k,v

    def popitem(self):
        return self.pop()

    def peek(self):
        return self.key[1], super().peek()

    cdef void _remove(self, int i):
        # update the locator stuff for last -> i
        last = self.val.end - 1
        self.swap(i, last)
        old = self.val.pop()
        # remove the key/loc/val associated with the deleted node.
        self.loc.pop(self.key.pop(last))
        # special handling for when the heap has size one.
        if i == last: return
        self._update(i, old, self.val.val[i])

    def __delitem__(self, k):
        self._remove(self.loc[k])

    def __contains__(self, k):
        return k in self.loc

    def __getitem__(self, k):
        return self.val.val[self.loc[k]]

    def __setitem__(self, k, v):
        self._setitem(k, v)

    cdef _setitem(self, object k, double v):
        "upsert (update or insert) value associated with key."
        cdef int i
        if k in self:
            # update
            i = self.loc[k]
            self._update(i, self.val[i], v)
        else:
            # insert (put new element last and bubble up)
            i = self.val.push(v)
            # Annoyingly, we have to write key/loc here the super class's push
            # method doesn't allow us to intervene before the up call.
            self.val[i] = v
            self.loc[k] = i
            self.key[i] = k
            # fix invariants
            self.up(i)

    cdef void swap(self, int i, int j):
        assert i < self.val.end
        assert j < self.val.end
        self.val.val[i], self.val.val[j] = self.val.val[j], self.val.val[i]

        self.key[i], self.key[j] = self.key[j], self.key[i]
        self.loc[self.key[i]] = i
        self.loc[self.key[j]] = j

    def check(self):
        super().check()
        for key in self.loc:
            assert self.key[self.loc[key]] == key
        for i in range(1, self.val.end):
            assert self.loc[self.key[i]] == i


class MinMaxHeap:

    def __init__(self, **kw):
        self.max = LocatorMaxHeap(**kw)
        self.min = LocatorMaxHeap(**kw)   # will pass negative values here

    def __contains__(self, k):
        return k in self.max

    def __setitem__(self, k, v):
        self.max[k] = v
        self.min[k] = -v

    def peekmin(self):
        k, v = self.min.peek()
        return k, -v

    def peekmax(self):
        return self.max.peek()

    def popmax(self):
        k, v = self.max.pop()
        del self.min[k]  # remove it from the min heap
        return k, v

    def popmin(self):
        k, v = self.min.pop()
        del self.max[k]  # remove it from the max heap
        return k, -v

    def check(self):
        self.min.check()
        self.max.check()

    def __len__(self):
        return len(self.max)

    def __repr__(self):
        return repr(self.max)

    def map(self):
        return {k: self.max[k] for k in self.max.loc}


class BoundedMaxHeap(MinMaxHeap):

    def __init__(self, maxsize, **kw):
        super().__init__(**kw)
        self.maxsize = maxsize

    def __setitem__(self, k, v):
        super().__setitem__(k, v)
        if len(self) > self.maxsize:
            if v < self.peekmin()[1]:  # smaller than the smallest element.
                return
            else:
                self.popmin()   # evict the smallest element

    def pop(self):
        return self.popmax()

    def check(self):
        super().check()
        assert len(self.max) <= self.maxsize
        assert len(self.min) <= self.maxsize



from bisect import bisect_left


def sorted_intersection(A, B):
    """Baeza-Yates set intersection, aka double-binary search.

    Let `N = min(|A|, |B|)` and `M=max(|A|, |B|)`

     - hash-based set intersection: O(N+M)
     - Baeza-Yates intersection: O(N log M) worst-case, better for most problems
       on "average"

    References:
    - https://mail.python.org/pipermail/python-list/2008-April/508321.html

    """

    if len(A) > len(B): A,B = B,A   # The smaller set drives the outer loop.

    def rec(lo1, hi1, lo2, hi2):
        if hi1 <= lo1: return
        if hi2 <= lo2: return

        mid1 = (lo1 + hi1) // 2
        x1 = A[mid1]                                # median of this part of set A

        mid2 = bisect_left(B, x1, lo=lo2, hi=hi2)   # find position of mid1 in B

        yield from rec(lo1, mid1, lo2, mid2)     # recurse on the first halves

        # Is the median of `A` in the intersection? (uses binary search above)
        if mid2 < hi2 and x1 == B[mid2]:
            #assert max(A[lo1], B[lo2]) <= x1 <= min(A[hi1-1], B[hi2-1])
            yield x1     # recursing in this order ensures that `out` is sorted.

        yield from rec(mid1+1, hi1, mid2, hi2)   # recurse on the second halves

    yield from rec(0, len(A), 0, len(B))


def test():
    import numpy as np
    import matplotlib.pyplot as pl
    from arsenal.timer import timers
    from arsenal import iterview

    T = timers()

    params = [
        (i, rep)
        for i in range(5, 20)
        for rep in range(5)
    ]

    np.random.shuffle(params)

    for (i, _) in iterview(params):

        # BY's runtime scales logarithmically with the bigger set and linearly
        # with the smaller set.  Hash-based intersetion scales with their sum.
        n = 2**i
        m = 2*10

        U = range(max(n,m)*5)
        A = list(np.random.choice(U, n, replace=0))
        B = list(np.random.choice(U, m, replace=0))
        A.sort()
        B.sort()

        sA = set(A)
        sB = set(B)
        with T['set'](i=2**i):
            E = (sA & sB)

        with T['make-set'](i=2**i):
            E = (set(A) & set(B))

        with T['B-Y'](i=2**i):
            C = list(sorted_intersection(A, B))

        assert sorted(E) == C

    T.compare()
    T.plot_feature('i')

    pl.show()


if __name__ == '__main__':
    test()



# TODO: This module should move into iterextras since it is focused on
# iterators, which are not specific to mathematics.


from itertools import product
from scipy.special import comb as binom, factorial
from arsenal import colors
import itertools



def sample(S, k, ordered, replace):
    S = tuple(S)

    if k == 0:
        yield ()
    else:
        for i, s in enumerate(S):

            if     ordered and     replace: X = S[:i] + S[i:]    # string
            if     ordered and not replace: X = S[:i] + S[i+1:]  # select
            if not ordered and     replace: X =         S[i:]    # ????
            if not ordered and not replace: X =         S[i+1:]  # choose

            for x in sample(X, k-1, ordered, replace):
                yield x + (s,)


def select(S, k):
    "Ordered subsets of size k"
    return sample(S, k, ordered=True, replace=False)


def choose(S, k):
    """
    Enumerate unordered subsets of size k from the set S.

    S does not have to be a set, it can be ordered.  The elements returned will each be a k-tuple.

    If `S` is an ordered collection, then the k-tuples returned by this iterator
    will use the sample order.

    """
    return sample(S, k, ordered=False, replace=False)


def permute(S):
    "Permutations"
    return select(S, len(S))


def n_selections_with_replacement(n, k):
    """
    Number of positive integers solutions to x₁ + x₂ + ⋯ + x_n = n+k
    (Linear diophantine equation)

    Also known as "((n choose k))" with doubled parentheses instead if single.

    https://www.johndcook.com/blog/select_with_replacement/

    """
    return binom(n+k-1, k)


def sumsto(n, k):
    assert n >= 0 and k >= 0
    if k == 0:
        return
    elif k == 1:
        yield (n,)
    else:
        for j in range(n+1):
            for js in sumsto(n-j, k-1):
                yield (j, *js)


# TODO: see itertools.{combinations combinations_with_replacement, permutations,
# product} those implementations are fair because they are based on "pools"

# TODO: implement "fair" versions of all methods (when they exist)

# TODO: implement set-partition enumeration

# TODO: figure out the fair enumerate extension and get all the special cases
# for free.  We know that a fair enumerator exists because the `string` is a
# superset of all other sets.

def sample(S, k, ordered, replace):
    S = tuple(S)

    if k == 0:
        yield ()
    else:
        for i, s in enumerate(S):

            if     ordered and     replace: X = S[:i] + S[i:]    # string
            if     ordered and not replace: X = S[:i] + S[i+1:]  # select
            if not ordered and     replace: X =         S[i:]    # ????
            if not ordered and not replace: X =         S[i+1:]  # choose

            for x in sample(X, k-1, ordered, replace):
                yield (s,) + x


def select(S, k):
    "Ordered subsets of size k"
    return sample(S, k, ordered=True, replace=False)


def choose(S, k):
    "Unordered subsets of size k"
    return sample(S, k, ordered=False, replace=False)


def permute(S):
    "Permutations"
    return select(S, len(S))


def perm_sign(p):
    if len(p) == 1: return True
    t = _inversions(p)
    return -1 if t % 2 else +1


try:
    from blist import sortedlist
except ImportError:
    pass
def _fast_inversions(p):
    n = len(p)
    total = 0
    S = sortedlist()
    for k in reversed(range(n)):
        pos = S.bisect_left(p[k])
        S.add(p[k])
        total += pos
    return total

def _slow_inversions(p):
    n = len(p)
    t = 0
    for i in range(n):
        for j in range(i+1, n):
            if p[i] > p[j]:
                t += 1
    return t

def _inversions(p):
    #assert _slow_inversions(p) == _fast_inversions(p)
    return _fast_inversions(p)


# Note: There is no way to fairly enumerate powerset when S is infinite
def powerset(S):
    "Powerset"
    # choose all subsets of size k from variables
    for k in range(len(S)+1):
        yield from choose(S, k)


# TODO: If we use a dovetailed Cartesian products, then we can support fair
# enumerate over infinite sequences S.
def string(S, k):
    "Strings of length k"
    return sample(S, k, ordered=True, replace=True)


# TODO: if we "parse" left-to-right (instead of top-down) then we can support
# fair enumeration over infinite sequences S.
def trees(S):
    "Generate all binary trees over sequence S."
    S = tuple(S)

    def _trees(i, k):
        if k-i == 1:
            yield S[i]
        else:
            for j in range(i, k):
                for y in _trees(i, j):
                    for z in _trees(j, k):
                        yield (y, z)

    return _trees(0, len(S))


def kleene(S, n=None):
    """
    Kleene closure of S: The set of strings over the alphabet `S` with the option
    to specify a maximum length `n` or leave as `None`.
    """

    def _kleene(n):
        if n is not None and n < 0: return
        yield ()
        for x in _kleene(n if n is None else n-1):
            for s in S:
                yield x + (s,)

    return _kleene(n)


def catalan(n):
    """Catalan numbers:
    `1, 1, 2, 5, 14, 42, 132, 429, 1430, 4862, 16796, 58786, 208012, 742900, ...`

    Number of ways to insert balanced parentheses in a sequence of `n+1` symbols.

      - for `n=1` there is 1 way: `(ab)`
      - for `n=2` there are 2 ways: `((ab)c) or (a(bc))`;
      - for `n=3` there are 5 ways: `((ab)(cd)), (((ab)c)d), ((a(bc))d), (a((bc)d)), (a(b(cd)))`.

    References:
    - https://en.wikipedia.org/wiki/Catalan_number
    - https://oeis.org/A000108

    """
    return int(binom(2*n, n)) // (n+1)


def enumerate_digraphs(n):
    "Enumerate all directed graphs over `n` nodes."
    import networkx as nx, numpy as np
    for e in product(*([[0,1]]*(n*n))):
        e = np.array(e).reshape((n, n))
        yield nx.from_numpy_matrix(e, create_using = nx.DiGraph)


# Less efficient version
#def sumsto(n, k):
#    assert n >= 0 and k >= 0
#    for js in product(range(n+1), repeat=k):
#        if sum(js) == n:
#            yield js


def test_sumsto():
    for n in range(8):
        for k in range(n+1):
            #print(f'\nn={n}, k={k}')

            N = 0
            for js in sumsto(n, k):
                N += 1
                #print(' ➥', js)
                assert sum(js) == n
                assert len(js) == k
                assert all(j >= 0 for j in js)

            have = N
            want = binom(n+k-1, n)
            #print('total', colors.mark(have==want), have, want)
            assert have == want


def test_sample():
    print('[sample]')

    def check(N,K):
        S = range(N)
        assert binom(N,K) == length(sample(S, K, ordered=0, replace=0))
        #A01 = sample(S, K, ordered=0, replace=1)   # ???
        assert binom(N,K) * factorial(K) == length(sample(S, K, ordered=1, replace=0))
        assert N**K == length(sample(S, K, ordered=1, replace=1))

    check(5,3)

    # Check corner cases
    check(5,5)
    check(5,0)


def test_kleene():
    print('[kleene]')
    from arsenal.iterextras import take

    assert (list(take(10, map(''.join, kleene('01'))))
            == ['', '0', '1', '00', '01', '10', '11', '000', '001', '010'])

    assert (list(map(''.join, kleene('01', n=3)))
            == ['', '0', '1', '00', '01', '10', '11', '000', '001', '010',
                '011', '100', '101', '110', '111'])

    assert (list(map(''.join, kleene('a', n=4)))
            == ['', 'a', 'aa', 'aaa', 'aaaa'])

    assert (list(map(''.join, kleene('', n=4)))
            == [''])

    assert length(powerset(range(5))) == 2**5
    assert length(permute(range(5))) == factorial(5)


def length(xs):
    return sum(1 for _ in xs)


def flatten(S):
    if not isinstance(S, (list, tuple)):
        yield S
    else:
        for x in S:
            yield from flatten(x)


def test_trees():

    _catalan = [1, 1, 2, 5, 14, 42, 132, 429, 1430, 4862, 16796, 58786,
                208012, 742900, 2674440]

    for n in range(1, len(_catalan)):
        assert _catalan[n] == catalan(n), [n, _catalan[n], catalan(n)]

    verbose = False

    def test(S):
        A = list(trees(S))
        if verbose: print(f'\ntrees of {S}')
        for x in A:
            if verbose: print('  ', x)
            # Check that flattening tree gives the seqence `S`.
            assert tuple(flatten(x)) == tuple(S)
        assert length(A) == catalan(n-1)

    for n in range(2, 10):
        print(f'[trees] n={n}' )
        test(range(n))


def segmentations(S):
    "Enumerate all segmentations of S"
    N = len(S)
    p = [[] for _ in range(N+1)]
    p[0] = [()]
    for t in range(1, N+1):
        for s in range(t):
            p[t] += [prefix + (tuple(S[s:t]),) for prefix in p[s]]
    return p[N]


def test_segementations():

    verbose = False

    def test(S):
        A = list(segmentations(S))
        if verbose: print(f'\nsegementations of {S}')
        for i, p in enumerate(sorted(A)):
            if verbose: print(f'  {i}: {p}')
            # Check that `p` is actually a segmentation of S.
            assert ''.join(''.join(s) for s in p) == ''.join(S), S
        # Check that we have the correct number
        assert len(A) == 2 ** (len(S) - 1)
        # Check for duplicates
        assert len(A) == len(set(A))

    for n in range(2, 10):
        print(f'[segmentations] n={n}')
        test(tuple(map(str,range(n))))


#class Collection:
#    def __init__(self, xs):
#        self.xs = list(xs)
#    def __hash__(self):
#        return hash(self.xs)
#    def __repr__(self):
#        return repr(self.xs)
#    def __eq__(self, other):
#        return self.xs == other.xs
#    def choose(self, k):
#        return Collection(choose(self.xs, k))
#    def permute(self):
#        return Collection(permute(self.xs))
#    def select(self, k):
#        return Collection(select(self.xs, k))
#    def strings(self, k):
#        return Collection(sample(self.xs, k, ordered=True, replace=True))
#    def misc(self, k):
#        return Collection(sample(self.xs, k, ordered=False, replace=True))
#

#print(Collection('abcd').choose(3))
#print(Collection('ab').strings(3))
#print(Collection('ab').misc(3))
#print(Collection('abcd').select(3))
#exit()


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



from getpass import getpass
try:
    import keyring
except ImportError:
    import warnings
    warnings.warn('install `keyring` to avoid typing password every time.')
    keyring = None


def set_password(service, user, pw=None):
    print('Set password:')
    if pw is None:
        pw = getpass()
        print('Type password again (for verification):')
        if pw != getpass():
            print('Error: passwords did not match. Try again.')
            set_password(service, user)
    if keyring: keyring.set_password(service, user, pw)
    return pw


def password(service, user):
    pw = None
    if keyring:
        pw = keyring.get_password(service, user)
    if pw is None:
        pw = set_password(service, user)
    return pw



#!/usr/bin/env python

import os
import sys
import pstats
import cProfile
from arsenal import colors
from contextlib import contextmanager


class profiler:

    def __init__(self, use='cprofile', filename='/tmp/out.prof'):
        self.use = use
        self.filename = filename

    def __enter__(self):
        if self.use == 'yep':   # pragma: no cover
            import yep
            self.prof = yep
            self.prof.start(self.filename)
        if self.use == 'cprofile':  # pragma: no cover
            #import cProfile
            self.prof = cProfile.Profile()
            self.prof.enable()
        return self

    def __exit__(self, *args, **kwargs):
        if self.use == 'yep':  # pragma: no cover
            self.prof.stop()
            print(colors.yellow % 'wrote: %s' % filename, '(use `google-pprof` to view)')
            # google-pprof --text /bin/ls imitation.prof
            # google-pprof --evince /bin/ls imitation.prof
            # google-pprof --web /bin/ls --web imitation.prof
        if self.use == 'cprofile':  # pragma: no cover
            #import pstats
            self.prof.disable()
            self.prof.dump_stats(self.filename)
            #pstats.Stats(filename).strip_dirs().sort_stats('time').print_stats()
            print(colors.yellow % 'wrote: %s' % self.filename, '(use `gprof-viz` to view)')

    def graphviz(self):
        if self.use == 'yep':  # pragma: no cover
            raise NotImplementedError()
        if self.use == 'cprofile':  # pragma: no cover
            return prof_to_graphviz(self.filename)

    def open(self):
        self.graphviz().view()


def prof_to_graphviz(f_prof):
    import gprof2dot
    from graphviz import Source

    f_dot = f_prof + '.dot'

    gprof2dot.main(['-f', 'pstats', '-o', f_dot, f_prof])

    with open(f_dot) as f:
        g = Source(f.read())

    os.remove(f_dot)    # cleanup

    return g


# TODO:
#  - maybe we should delete the `out` tempfile
#  - I don't like that this currently makes a system call to open the image...
#  - gprof2dot is on the pythonpath we should be able to avoid the system call
def profile_viz(
        cmd, global_dict=None,
        local_dict=None,
        img='/tmp/profile.png',
        out='/tmp/profile.tmp',
        noctx=False
):
    "Run gprof2dot on the output for profiling."
    if noctx:
        cProfile.run(cmd, out)
    else:
        if local_dict is None and global_dict is None:
            call_frame = sys._getframe().f_back
            local_dict = call_frame.f_locals
            global_dict = call_frame.f_globals
        cProfile.runctx(cmd, global_dict, local_dict, out)

    stats = pstats.Stats(out)
    stats.strip_dirs()               # Clean up filenames for the report
    stats.sort_stats('cumulative')   # sort by the cumulative time
    stats.print_stats()
    # for more on the viz check out: http://code.google.com/p/jrfonseca/wiki/Gprof2Dot
    os.system('gprof2dot.py -f pstats %s | dot -Tpng -o %s && eog %s &' % (out, img, img))


#def kcachegrind(cmd, out='profile.kgrind'):
#    from arsenal.profiling.lsprofcalltree import KCacheGrind
#    p = cProfile.Profile()
#    p.run(cmd)
#    # Get the stats in a form kcachegrind can use and save it
#    k = KCacheGrind(p)
#    with file(out, 'wb') as f:
#        k.output(f)
#    os.system("kcachegrind %s &" % out)


#def main():
#    from optparse import OptionParser
#    parser = OptionParser()
#    parser.allow_interspersed_args = False
#    parser.add_option('-o', '--outfile', dest="outfile",
#                      help="Save stats to <outfile>",
#                      default='/tmp/profile.tmp')
#
#    if not sys.argv[1:]:
#        parser.print_usage()
#        sys.exit(2)
#
#    (options, args) = parser.parse_args()
#
#    #viz = kcachegrind
#    viz = profile_viz
#
#    sys.path = [os.getcwd()] + sys.path
#
#    if len(args) > 0:
#        sys.argv[:] = args
#        sys.path.insert(0, os.path.dirname(sys.argv[0]))
#        viz('execfile(%r)' % sys.argv[0], out=options.outfile)
#    else:
#        parser.print_usage()
#    return parser
#
#
#if __name__ == '__main__':
#    main()


def test_profiler():
    import time

    with profiler() as p:
        for i in range(10):
            time.sleep(.1)

    p.open()


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



class BucketQueue:
    """BucketQueue.py
    Taken from https://github.com/timvieira/dyna-pi/blob/main/dyna/util/bucket_queue.py

    A simple priority queue with integer priorities. Use like:
        Q = BucketQueue()
        Q[item] = priority
        del Q[item]
        for item in Q: ...

    The time to find and return each item in the queue is proportional
    to the difference (current priority - previous priority), or O(1) if
    this difference is non-positive. In particular when the minimum priority
    is non-negative and non-decreasing (as in Dijkstra's algorithm) or when
    its decreases are bounded (as in graph degeneracy) the total time for a
    sequence of operations is proportional to the number of operations plus
    the maximum priority.

    D. Eppstein, July 2016.

    References:
    * https://www.ics.uci.edu/~eppstein/261/s21w5.pdf

    """
    def __init__(self):
        """Create a new empty integer priority queue."""
        self._D = {}        # map from items to priorities
        self._Q = {}        # map from priorities to buckets
        self._N = None      # lower bound on min priority

    def __getitem__(self, item):
        "Look up the priority of an item."
        return self._D[item]

    def __delitem__(self, item):
        "Remove an item from the priority queue."
        priority = self._D[item]
        del self._D[item]               # remove from map of items => priorities
        self._Q[priority].remove(item)  # remove from bucket
        if not self._Q[priority]:
            del self._Q[priority]       # remove empty bucket

    def __setitem__(self, item, priority):
        "Add an element to the priority queue with the given priority."
        if not isinstance(priority, int):
            raise TypeError("Priority must be an integer")
        if item in self._D:
            del self[item]
        self._D[item] = priority        # add to map of items => priorities
        if not self._Q or priority < self._N:
            self._N = priority          # update priority lower bound
        if priority not in self._Q:
            self._Q[priority] = set()   # make new bucket if necessary
        self._Q[priority].add(item)     # and add to bucket

    def __iter__(self):
        """Repeatedly find and remove the min-priority item from the queue.
        It is ok for the queue to be modified between iterations."""
        while self._Q:
            while self._N not in self._Q:
                self._N += 1
            x = next(iter(self._Q[self._N]))    # arbitrary item in 1st bucket
            del self[x]
            yield x

    def popitem(self):
        while self._Q:
            while self._N not in self._Q:
                self._N += 1
            x = next(iter(self._Q[self._N]))    # arbitrary item in 1st bucket
            del self[x]
            return (x, self._N)

    def pop(self):
        while self._Q:
            while self._N not in self._Q:
                self._N += 1
            x = next(iter(self._Q[self._N]))    # arbitrary item in 1st bucket
            del self[x]
            return x

    def items(self):
        """Variant iterator that generates (item,priority) pairs.
        We rely on the fact that the usual __iter__ always
        leaves self._N equal to the priority."""
        for x in self:
            yield (x, self._N)

    def __contains__(self, item):
        "Container class membership test."
        return item in self._D

    def __len__(self):
        "Container class length."
        return len(self._D)



"""
Paired permutation test for corpus statistics (non-additive).
"""
import numpy as np
from arsenal.iterview import iterview
from arsenal import colors


def paired_permutation_test(xs, ys, statistic, threshold=0.05, R=10_000, verbose=1):
    "Pair permutation test"

    def effect(xs,ys): return np.abs(statistic(xs) - statistic(ys))   # two-sided: A != B

    ra = statistic(xs)
    rb = statistic(ys)
    diff = effect(xs,ys)       # observed difference

    n = len(xs)
    k = 0
    reps = range(R)
    if verbose: reps = iterview(reps, msg='perm test')
    for _ in reps:
        # randomly generate a vector of zeros and ones (uniformly).
        swaps = np.random.randint(0,2,n).astype(bool)          # flip n coins
        k += diff <= effect(np.select([swaps,~swaps],[xs,ys]), # swap elements accordingly
                            np.select([~swaps,swaps],[xs,ys]))

    s = k/R

    #threshold = 0.5*threshold
    #s *= 0.5   # because we have a two-sided test.

    if verbose:
        # which system has higher reward? is it significant?
        asig = (colors.red % colors.bold) if ra > rb and s <= threshold else '%s'
        bsig = (colors.blue % colors.bold) if rb > ra and s <= threshold else '%s'
        any_sig = colors.bold if s <= threshold else colors.yellow

        print(asig % f'R(A) = {ra:g}')
        print(bsig % f'R(B) = {rb:g}')
        print(any_sig % f'confidence = {(1-s)}\n')

    if s <= threshold:
        return s, -1 if ra > rb else +1
    else:
        return s, 0   # "statistical tie"



def test_f1():

    if 0:
        def f1(correct, predict, n_labels):
            assert correct.shape == predict.shape
            tp, fp, fn = np.zeros((3, n_labels))
            for (c, p) in zip(correct, predict):
                if c == p:
                    tp[c] += 1
                else:
                    fp[p] += 1
                    fn[c] += 1
            F,P,R = np.zeros((3, n_labels))
            for c in range(n_labels):
                if tp[c] > 0:
                    P[c] = tp[c] / (tp[c] + fp[c])
                    R[c] = tp[c] / (tp[c] + fn[c])
                    F[c] = 2*P[c]*R[c] / (P[c] + R[c])
            return F,P,R

        def test_statistic(predict):
            f,_,_ = f1(c, predict, 2)
            return f[1]   # F1-measure for label 1.
    else:
        # accuracy
        def test_statistic(predict): return np.mean(c == predict)

    t = 0.05
    c = np.array([0,1,0,1,0,1,0,1,0,1,0,1])   # correct labels
    a = np.array([1,1,1,1,0,1,1,0,1,1,0,0])
    b = np.array([0,1,0,1,0,1,0,1,0,1,0,1])
    paired_permutation_test(a, b, statistic=test_statistic,
                            threshold=t, verbose=True)


if __name__ == '__main__':
    test_f1()



"""
Pareto frontier
"""
import matplotlib.pyplot as pl
import numpy as np
from arsenal import colors
from arsenal.iterextras import window


def pareto_frontier(X, Y, maxX=True, maxY=True, indices=False):
    """Determine Pareto frontier, returns list of sorted points.

    Args:

      X, Y: data.

      maxX, maxY: (bool) whether to maximize or minimize along respective
        coordinate.

    """
    assert len(X) == len(Y)
    if len(X) == 0:
        return []

    xx = -1 if maxX else +1
    yy = -1 if maxY else +1
    key = lambda xy: (xy[0] * xx, xy[1] * yy)   # need to break ties

    a = sorted(zip(X, Y, list(range(len(X)))), key=key)
    frontier = []
    lastx = float('inf') * xx
    lasty = float('inf') * yy
    for xy in a:
        x,y,i = xy
        if yy*y <= yy*lasty:
            if lastx != x:
                frontier.append(xy)
            lasty = y
            lastx = x

    x,y,i = list(zip(*frontier))
    return list(i) if indices else list(zip(x,y))


def pareto_ix(X, Y, *a, **kw):
    "Determine Pareto frontier, returns list of indices"
    return pareto_frontier(X, Y, *a, **kw)
#    assert len(X) == len(Y)
#    if len(X) == 0:
#        return []
#    a = sorted(zip(X, Y, range(len(X))), reverse=maxX)
#    frontier = []
#    lastx = float('-inf') if maxX else float('+inf')
#    lasty = float('-inf') if maxY else float('+inf')
#    for xyi in a:
#        x,y,i = xyi
#        if maxY:
#            if y >= lasty:
#                if lastx != x:
#                    frontier.append(i)
#                lasty = y
#                lastx = x
#        else:
#            if y <= lasty:
#                if lastx != x:
#                    frontier.append(i)
#                lasty = y
#                lastx = x
#    return frontier


def show_frontier(X, Y, maxX=False, maxY=True, dots=False,
                  XMIN=None, XMAX=None,
                  YMIN=None, YMAX=None,
                  ax=None, label=None,
                  interpolation='pessimistic', **style):
    """Plot Pareto frontier.

    Args:

      X, Y: data.

      maxX, maxY: (bool) whether to maximize or minimize along respective
        coordinate.

      dots: (bool) highlight points on the frontier (will use same color as
        `style`).

      ax: use an existing axis if non-null.

      style: keyword arguments, which will be passed to lines connecting the
        points on the Pareto frontier.

      XMAX: max value along x-axis
      YMIN: min value along y-axis

    """
    if ax is None:
        ax = pl.gca()
    sty = {'c': 'b', 'alpha': 0.3, 'zorder': 0}
    sty.update(style)

    if interpolation == 'linear-convex':
        # Convex hull by itself doesn't work, but what we have is ok because its
        # the intersection of the convex hull with the pareto frontier, which is
        # handled below.
        from scipy.spatial import ConvexHull
        X = np.array(X)
        Y = np.array(Y)
        hull = ConvexHull(np.array([X,Y]).T)
        X = X[hull.vertices]
        Y = Y[hull.vertices]

#    assert not maxX and maxY, 'need to update some hardcoded logic'

    f = pareto_frontier(X, Y, maxX=maxX, maxY=maxY)
    if not f:
        print(colors.yellow % '[warn] Empty frontier')
        return
    if dots:
        xs, ys = list(zip(*f))
        ax.scatter(xs, ys, lw=0, alpha=0.5, c=sty['c'])

    XMIN = min(min(X), XMIN) if XMIN is not None else min(X)
    XMAX = max(max(X), XMAX) if XMAX is not None else max(X)
    YMIN = min(min(Y), YMIN) if YMIN is not None else min(Y)
    YMAX = max(max(Y), YMAX) if YMAX is not None else max(Y)

#    if 0:
#        # This version lets you clip away points. I found it a little problematic to use.
#        XMIN = XMIN if XMIN is not None else min(X)
#        XMAX = XMAX if XMAX is not None else max(X)
#        YMIN = YMIN if YMIN is not None else min(Y)
#        YMAX = YMAX if YMAX is not None else max(Y)
#        if (max(X) > XMAX or min(X) < XMIN or min(Y) > YMIN or max(Y) < YMAX):
#            print '[PARETO] Warning: data out of bounds!'

    # Connect corners of frontier. The first and last points on frontier have
    # lines which surround the point cloud.

    # TODO: make this look nicer...
    if maxX and maxY:
        f = [(XMAX, YMIN)] + f + [(XMIN, YMAX)]
    elif not maxX and maxY:
        f = [(XMIN, YMIN)] + f + [(XMAX, YMAX)]
    elif maxX and not maxY:
        f = [(XMAX, YMAX)] + f + [(XMIN, YMIN)]
    else:
        f = [(XMIN, YMAX)] + f + [(XMAX, YMIN)]

    if interpolation == 'pessimistic':
        # Make line segments from adjacent points
        pts = np.array([x for ((a,b), (c,d)) in window(f, 2) for x in [[a,b], [c,b], [c,b], [c,d]]])
    elif interpolation in {'linear','linear-convex'}:
        # Make line segments from adjacent points
        pts = np.array([x for ((a,b), (c,d)) in window(f, 2) for x in [[a,b], [c,d]]])

    # Plot
    ax.plot(pts[:,0], pts[:,1], label=label, **sty)
    return pts


class Pareto(object):

    def __init__(self, df, xcol, ycol, ax=None):
        self.df = df
        self.xcol = xcol
        self.ycol = ycol
        self.ax = ax
        self.frontier = pareto_ix(df[xcol], df[ycol])
        assert np.isfinite(df[xcol]).all() and np.isfinite(df[ycol]).all(), \
            'Pareto: `DataFrame` contains non-finite values.'

    def scatter(self, **plot_kwargs):
        if self.ax is None:
            self.ax = pl.gca()
        kwargs = dict(lw=0)
        kwargs.update(plot_kwargs)
        self.ax.scatter(self.df[self.xcol], self.df[self.ycol], **kwargs)

    def show_frontier(self, **plot_kwargs):
        if self.ax is None:
            self.ax = pl.gca()
        show_frontier(self.df[self.xcol], self.df[self.ycol], ax=self.ax, **plot_kwargs)

    def lookup_x(self, x):
        """
        Find Pareto point constrained by x.

          argmax       p.y
          p: p.x <= x

        """
        s = self.df
        s = s[s[self.xcol] <= x]    # filter
        if s.empty:
            return np.nan
        yy = s[self.ycol].argmax()
        return s.ix[yy][self.ycol]

    def lookup_y(self, y):
        """
        Find Pareto point constrained by y.

          argmax    p.x
          p: p.x >= y

        """
        s = self.df
        s = s[s[self.ycol] >= y]    # filter
        if s.empty:
            return np.nan
        xx = s[self.xcol].argmin()
        return s.ix[xx][self.xcol]


def test():
    if 0:
        from pandas import DataFrame
        X = np.linspace(0.01, 1.0, 10)
        Y = np.log(X)
        Y -= Y.min()
        Y /= Y.max()
        Y *= 0.95

        df = DataFrame({'X': X, 'Y': Y})
        P = Pareto(df, 'X', 'Y')

        data = []
        for val in np.linspace(0,1,15):
            data.append(dict(val=val, x=P.lookup_x(val), y=P.lookup_y(val)))
            pl.axvline(val, alpha=.5)
            pl.axhline(val, alpha=.5)
        dd = DataFrame(data)
        pl.scatter(dd.y, dd.val, lw=0, c='r')
        pl.scatter(dd.val, dd.x, lw=0, c='g')
        print(dd)

        #P.scatter(c='r', lw=0)
        P.show_frontier(c='r', lw=4)
        pl.show()

    X,Y = np.random.normal(0,1,size=(2, 30))

    for maxX in [0,1]:
        for maxY in [0,1]:
            pl.figure()
            pl.title('max x: %s, max y: %s' % (maxX, maxY))
            pl.scatter(X,Y,lw=0)
            show_frontier(X, Y, maxX=maxX, maxY=maxY)
            pl.show()


if __name__ == '__main__':
    test()



# cython: language_level=3, boundscheck=False, infer_types=True, nonecheck=False
# cython: overflowcheck=False, initializedcheck=False, wraparound=False, cdivision=True
import numpy as np

from libc.stdlib cimport rand
from libc.math cimport log2, ceil
cdef extern from "limits.h":
    int INT_MAX

cdef double MAX = float(INT_MAX)
cdef inline double uniform() nogil:
    return rand() / MAX


cdef class SumHeap:
    cdef readonly:
        double[:] S
        int n, d

    def __init__(self, double[:] w):
        self.n = w.shape[0]
        self.d = int(2**ceil(log2(self.n)))   # number of intermediates
        self.S = np.zeros(2*self.d)           # intermediates + leaves
        self.heapify(w)

    def __getitem__(self, int k):
        return self.S[self.d + k]

    def __setitem__(self, int k, double v):
        self.update(k, v)

    cpdef void heapify(self, double[:] w):
        "Create sumheap from weights `w` in O(n) time."
        d = self.d; n = self.n
        self.S[d:d+n] = w                         # store `w` at leaves.
        for i in reversed(range(1, d)):
            self.S[i] = self.S[2*i] + self.S[2*i + 1]

    cpdef void update(self, int k, double v):
        "Update w[k] = v` in time O(log n)."
        i = self.d + k
        self.S[i] = v
        while i > 0:   # fix parents in the tree.
            i //= 2
            self.S[i] = self.S[2*i] + self.S[2*i + 1]

    cpdef int sample(self, u=None):
        "Sample from sumheap, O(log n) per sample."
        cdef double left, p
        if u is None: u = uniform()
        d = self.S.shape[0]//2     # number of internal nodes.
        p = u * self.S[1]  # random probe, p ~ Uniform(0, z)
        # Use binary search to find the index of the largest CDF (represented as a
        # heap) value that is less than a random probe.
        i = 1
        while i < d:
            # Determine if the value is in the left or right subtree.
            i *= 2            # Point at left child
            left = self.S[i]  # Probability mass under left subtree.
            if p > left:      # Value is in right subtree.
                p -= left     # Subtract mass from left subtree
                i += 1        # Point at right child
        return i - d

    cpdef long[:] swor(self, int k):
        "Sample without replacement `k` times."
        cdef long[:] z = np.zeros(k, dtype=int)
        for i in range(k):
            k = self.sample()
            z[i] = k
            self.update(k, 0)
        return z



import matplotlib.pyplot as pl
import numpy as np
import pandas as pd
from collections import defaultdict


class Measurements:
    def __init__(self, colors):
        self.data = {name: defaultdict(list) for name in colors}
        self.colors = colors

    def __call__(self, name, x, samples):
        samples = np.array(samples)
        avg = samples.mean()
        self.data[name]['xs'].append(x)
        self.data[name]['avg'].append(avg)
        if 0:
            std = samples.std()
            hi = avg+std; lo = avg-std
        else:
            hi = np.percentile(samples, 80)
            lo = np.percentile(samples, 20)
        self.data[name]['hi'].append(hi)
        self.data[name]['lo'].append(lo)
        #self.data[name]['samples'].append(samples)

    def show(self, ax=None):
        if ax is None: ax = pl.figure().add_subplot(111)
        # plot all methods +/- error bars.
        for name, data in self.data.items():
            if not data: continue
            df = pd.DataFrame(data)
            pl.fill_between(df['xs'], df.hi, df.lo, alpha=0.1, color=self.colors[name])
            pl.plot(df['xs'], df['avg'], lw=2, c=self.colors[name], alpha=0.5, label=name)
        pl.legend(loc=4, ncol=len(self.data), borderaxespad=0.0, fontsize=10, frameon=0,
                  columnspacing=0.5, handletextpad=0.0)

    def clear(self):
        for d in self.data.values():
            d.clear()



from datetime import datetime, timedelta
from time import time, localtime, strftime
import atexit


# TODO: option to use terminal width if possible
def marquee(txt='', width=78, mark='*'):
    """
    Return the input string centered in a 'marquee'.

    >>> marquee('hello', width=50)
    '********************* hello *********************'
    """
    if not txt:
        return (mark*width)[:width]
    nmark = int((width-len(txt)-2)/len(mark)/2)
    if nmark < 0: nmark =0
    marks = mark*nmark
    return '%s %s %s' % (marks, txt, marks)


def nth(n):
    """
    Formats an ordinal.
    Doesn't handle negative numbers.

    >>> nth(1)
    '1st'
    >>> nth(0)
    '0th'
    >>> [nth(x) for x in [2, 3, 4, 5, 10, 11, 12, 13, 14]]
    ['2nd', '3rd', '4th', '5th', '10th', '11th', '12th', '13th', '14th']
    >>> [nth(x) for x in [91, 92, 93, 94, 99, 100, 101, 102]]
    ['91st', '92nd', '93rd', '94th', '99th', '100th', '101st', '102nd']
    >>> [nth(x) for x in [111, 112, 113, 114, 115]]
    ['111th', '112th', '113th', '114th', '115th']
    """
    assert n >= 0
    if n % 100 in [11, 12, 13]:
        return '%sth' % n
    return {1: '%sst', 2: '%snd', 3: '%srd'}.get(n % 10, '%sth') % n


def timetuple(s):
    "htime(x) -> (days, hours, minutes, seconds)"
    s = int(s)
    d, s = divmod(s, 86400)
    h, s = divmod(s, 3600)
    m, s = divmod(s, 60)
    return (d, h, m, s)


def htime(s, show_seconds=True):
    """Given a number of seconds, returns a string attempting to represent
    it as shortly as possible.
    >>> htime(100000)
    '1d3h46m40s'
    >>> htime(1)
    '1s'
    >>> htime(10)
    '10s'
    >>> htime(1000)
    '16m40s'
    >>> htime(10000)
    '2h46m40s'
    """
    (d, h, m, s) = timetuple(s)
    x = []
    if d:
        x.append('%sd' % d)
    if h:
        x.append('%sh' % h)
    if m:
        x.append('%sm' % m)
    if show_seconds and s:
        x.append('%ss' % s)
    if not x:
        if show_seconds:
            x = ['%ss' % s]
        else:
            x = ['0m']
    return ''.join(x)


def datestr(then, now=None):
    """
    Converts a (UTC) datetime object to a nice string representation.

    >>> d = datetime(1970, 5, 1)
    >>> datestr(d, now=d)
    '0 microseconds ago'
    >>> datestr(datetime(1970, 1, 1), now=d)
    'Jan 01'
    >>> datestr(datetime(1969, 1, 1), now=d)
    'Jan 01, 1969'
    >>> datestr(datetime(1970, 6, 1), now=d)
    'Jun 01, 1970'
    >>> datestr(None)
    ''
    """

    def agohence(n, what, divisor=None):
        if divisor:
            n //= divisor
        out = str(abs(n)) + ' ' + what       # '2 day'
        if abs(n) != 1:
            out += 's'                       # '2 days'
        out += ' '                           # '2 days '
        if n < 0:
            out += 'from now'
        else:
            out += 'ago'
        return out                           # '2 days ago'

    oneday = 24 * 60 * 60

    if not then:
        return ""
    if not now:
        now = datetime.now()

    delta = now - then
    deltaseconds = int(delta.total_seconds())
    deltadays = abs(deltaseconds) // oneday
    if deltaseconds < 0:
        deltadays *= -1 # fix for oddity of floor

    if deltadays:
        if abs(deltadays) < 4:
            return agohence(deltadays, 'day')

        out = then.strftime('%b %d') # e.g. 'June 13'
        if then.year != now.year or deltadays < 0:
            out += ', %s' % then.year
        return out

    if int(deltaseconds):
        if abs(deltaseconds) > (60 * 60):
            return agohence(deltaseconds, 'hour', 60 * 60)
        elif abs(deltaseconds) > 60:
            return agohence(deltaseconds, 'minute', 60)
        else:
            return agohence(deltaseconds, 'second')

    deltamicroseconds = delta.microseconds
    if delta.days:
        deltamicroseconds = int(delta.microseconds - 1e6) # datetime oddity
    if abs(deltamicroseconds) > 1000:
        return agohence(deltamicroseconds, 'millisecond', 1000)

    return agohence(deltamicroseconds, 'microsecond')


# TODO: use htime and marquee
def print_elapsed_time():
    "register an exit hook which prints the start, finish, and elapsed times of a script."
    begin = time()
    started = localtime()
    def hook():
        secs = time() - begin
        mins, secs = divmod(secs, 60)
        hrs, mins = divmod(mins, 60)
        print()
        print('======================================================================')
        print('Started on', strftime("%B %d, %Y at %I:%M:%S %p", started))
        print('Finished on', strftime("%B %d, %Y at %I:%M:%S %p", localtime()))
        print('Total time: %02d:%02d:%02d' % (hrs, mins, secs))
        print()
    atexit.register(hook)


# borrowed from girzzled https://github.com/bmc/grizzled-python/blob/master/grizzled/text/__init__.py
def str2bool(s):
    """
    Convert a string to a boolean value. The supported conversions are:

        +--------------+---------------+
        | String       | Boolean value |
        +==============+===============+
        | "false"      | False         |
        +--------------+---------------+
        | "true"       | True          |
        +--------------+---------------+
        | "f"          | False         |
        +--------------+---------------+
        | "t"          + True          |
        +--------------+---------------+
        | "0"          | False         |
        +--------------+---------------+
        | "1"          + True          |
        +--------------+---------------+
        | "n"          | False         |
        +--------------+---------------+
        | "y"          + True          |
        +--------------+---------------+
        | "no"         | False         |
        +--------------+---------------+
        | "yes"        + True          |
        +--------------+---------------+
        | "off"        | False         |
        +--------------+---------------+
        | "on"         + True          |
        +--------------+---------------+

    Strings are compared in a case-blind fashion.

    **Note**: This function is not currently localizable.

    :Parameters:
        s : str
            The string to convert to boolean

    :rtype: bool
    :return: the corresponding boolean value

    :raise ValueError: unrecognized boolean string
    """
    try:
        return {
            'true' : True,  'false': False,
            't'    : True,  'f'    : False,
            '1'    : True,  '0'    : False,
            'yes'  : True,  'no'   : False,
            'y'    : True,  'n'    : False,
            'on'   : True,  'off'  : False,
        }[s.lower()]
    except KeyError:
        raise ValueError('Unrecognized boolean string: "%s"' % s)


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    def test():
        d = datetime(1970, 5, 1)

        examples = {
            timedelta(microseconds=1): '1 microsecond ago',
            timedelta(microseconds=2): '2 microseconds ago',
            -timedelta(microseconds=1): '1 microsecond from now',
            -timedelta(microseconds=2): '2 microseconds from now',
            timedelta(microseconds=2000): '2 milliseconds ago',
            timedelta(seconds=2): '2 seconds ago',
            timedelta(seconds=2*60): '2 minutes ago',
            timedelta(seconds=2*60*60): '2 hours ago',
            timedelta(days=2): '2 days ago',
        }

        for t, v in list(examples.items()):
            assert datestr(d, now=d+t) == v

    test()



# TODO:
#  - consider allowing lazy things to depend on one another

from types import GeneratorType

class lazy(object):
    """
    Lazily load a property defined by a method. The method wrapped is called
    at most once to retrieve the result and the result is reused. If the method
    is a generator, the value is stored as a list.

    Note: instances must have a `__dict__` attribute in order for this property
    to work, i.e. no '__slots__' class attribute.

    Implementation detail: this is not implemented as a data descriptor so that
    we can completely avoid the function call overhead. If one choses to invoke
    `__get__` by hand the property will still work as expected because the
    lookup logic is replicated in `__get__` for manual invocation.
    """

    def __init__(self, func):
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.func = func

    def __get__(self, obj, type_=None):
        if obj is None:
            return self
        try:
            value = obj.__dict__[self.__name__]
        except KeyError:
            value = self.func(obj)
            if isinstance(value, GeneratorType):   # store generators as lists
                value = list(value)
            obj.__dict__[self.__name__] = value
        return value

    def __set__(self, obj, value):
        raise NotImplementedError

    def __delete__(self, obj):
        raise NotImplementedError


if __name__ == '__main__':
    from collections import defaultdict

    class Foo(object):
        def __init__(self, x):
            self.x = x
            self.log = defaultdict(int)       # track the number of calls to our cached properties/methods
        @lazy
        def my_ondemand(self):
            self.log['ondemand'] += 1
            return 'ON DEMAND'.split()        # return something mutable
        @lazy
        def my_cached(self):
            self.log['cachedproperty'] += 1
            return 'CACHED PROPERTY'.split()
        @lazy
        def my_lazy(self):
            self.log['my_lazy'] += 1
            return 'LAZY PROPERTY'.split()
        @lazy
        def my_lazy_list(self):
            self.log['my_lazy_list'] += 1
            for i in range(10):
                yield i

    def test():
        import pickle

        def checks(x):
            o = x.my_ondemand
            c = x.my_cached
            l = x.my_lazy
            assert x.my_ondemand is o
            assert x.my_cached is c
            assert x.my_lazy is l
            print('x.log:', x.log, sep=' ')
            assert min(x.log.values()) == max(x.log.values()) == 1

            ll = x.my_lazy_list
            assert ll == list(range(10)), ll
            assert ll is x.my_lazy_list

            print(x.log)
            assert all(v == 1 for v in x.log.values())

        foo = Foo('XXX')
        checks(foo)
        foo.log.clear()

        print('pickling and unpickling..')
        foo2 = pickle.loads(pickle.dumps(foo))

        # should still call both lazy properties again.
        assert not foo2.log

    test()



import os
import inspect


def find_filename(obj, verbose=False):
    """Return the full filename of an instance. """

    if verbose:
        print('[find_filename] %r' % (obj,))

    if not isinstance(obj, str):

        # Return the name of the Python source file in which an object was defined.
        # This will fail with a TypeError if the object is a built-in module, class,
        # or function.
        try:
            f = inspect.getsourcefile(obj)
        except TypeError as e:
            if verbose:
                print('[find_filename] inspect.getsourcefile raised TypeError:', e)
            return None

    if verbose:
        print('[find_filename] found filename', f)

    # don't visit *.pyc files
    f = os.path.splitext(f)
    if f[1] == ".pyc":
        f = (f[0], ".py")
    f = ''.join(f)

    if not os.path.isabs(f):
        f = os.path.join(os.getcwd(), f)

    if os.path.isfile(f):
        return f


def edit(obj, verbose=True):
    """
    Set the synchronize with editor hook with a callable object.
     - obj: introspection is used to retrieve relevant source code for the
       object; strings are treated as a filenames.
     - line : the line number to scroll to.
     - column : the column number to scroll to.
    """

    filename = find_filename(obj, verbose=verbose)
    if not filename:
        if verbose:
            print('[edit] no file found for %r' % (obj,))
        return

    try:
        # Return a list of source lines and starting line number for an object.
        # The argument may be a module, class, method, function, traceback, frame,
        # or code object. The source code is returned as a list of the lines
        # corresponding to the object and the line number indicates where in the
        # original source file the first line of code was found. An IOError is raised
        # if the source code cannot be retrieved.
        _, line = inspect.getsourcelines(obj)
    except IOError:
        line = 0

    emacs(filename, line=line)


def emacs(filename, line=0, column=0):
    # emacsclient -n +%d:%d "%s" 2>/dev/null
    os.system('visit +%d:%d "%s" 2>/dev/null' % (line, column, filename))


if __name__ == '__main__':

    # TODO: look at misc.find_object
    import sys
    X = {}
    if len(sys.argv) == 2:
        exec('import %s as THING' % sys.argv[1], X)
    elif len(sys.argv) == 3:

        grab = '.'.join(['THING'] + sys.argv[2].split('.')[1:])
        exec("""from %s import %s as THING; THING=%s""" % (sys.argv[1], sys.argv[2].split('.')[0], grab), X)

    else:
        print('%s <module> or %s <module> <object>' % (sys.argv[0], sys.argv[0]))
        sys.exit(1)
    edit(X['THING'])



"""UnionFind.py

Union-find data structure. Based on Josiah Carlson's code,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
with significant additional changes by D. Eppstein.
"""

from collections import defaultdict

class UnionFind:
    """
    Union-Find data structure.

    Each `UnionFind` instance `X` maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - `X[item]` returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in `X`, a new singleton set is
      created for it.

    - `X.union(item1, item2, ...)` merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in `X`, it is added to `X` as one of the members of the merged set.
    """

    def __init__(self, elements=None):
        """Create a new empty union-find structure."""
        self.weights = {}
        self.parents = {}
        if elements is not None:
            for x in elements:
                self.add(x)

    @property
    def elems(self):
        return self.parents

    def add(self, x):
        "Add element x as a singleton"
        self.union(x, x)

    def connected(self, x, y):
        return self[x] == self[y]

    def __getitem__(self, obj):
        "Find and return the name of the set containing the object."

        # check for previously unknown object
        if obj not in self.parents:
            self.parents[obj] = obj
            self.weights[obj] = 1
            return obj

        # find path of objects leading to the root
        path = [obj]
        root = self.parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def __iter__(self):
        "Iterate through all items ever found or unioned by this structure."
        return iter(self.parents)

    def union(self, *objects):
        "Find the sets containing the objects and merge them all."
        roots = [self[x] for x in objects]
        heaviest = max(roots, key = self.weights.__getitem__)
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest

    def roots(self):
        for x in self:
            if self[x] == x:
                yield x

    def classes(self):
        classes = defaultdict(set)
        for x in self:
            root = self[x]  # does path compression as a side effect
            classes[root].add(x)
        return classes.values()

    def class_of(self, x):
        root = self[x]
        return [y for y in self if self[y] == root]



import numpy as np
from arsenal.maths.compare import compare
from arsenal.maths.rvs import spherical
from arsenal.iterview import iterview


# cg: Solve a linear systems, A x = b, using only A-x products. Assumes the
# matrix `A` is positive definite (same conditions as conjugate gradient). The
# cg solver requires that the Ax operator is an instance of LinearOperator
from scipy.sparse.linalg import cg as implicit_cg
from scipy.sparse.linalg import LinearOperator


from scipy.optimize import minimize
from scipy.linalg import norm
from arsenal import iterview


def fd_Hessian(f, x, eps=1e-5):
    [d] = x.flatten().shape
    x = x.flat
    [F] = f().shape
    H = np.zeros((d,d,F))
    for i in range(d):
        for j in range(d):
            # central-difference approximation of Hessian

            x_i = x[i]
            x_j = x[j]

            x[i] = x_i + eps
            x[j] = x_j + eps

            bb = f()

            x[i] = x_i + eps
            x[j] = x_j - eps

            ba = f()

            x[i] = x_i - eps
            x[j] = x_j + eps

            ab = f()

            x[i] = x_i - eps
            x[j] = x_j - eps

            aa = f()

            # reset
            x[i] = x_i
            x[j] = x_j

            H[i,j] = (bb - ba - ab + aa) / (4*eps**2)

    return H


def prox_numerical(f, x, s, jac=None):
    "Numerically estimate the proximal operator `Prox_{s*f}(x)`."

    def fg(z):
        d = (z-x)
        p = 0.5/s * (d @ d)
        if jac:
            fz, gz = f(z)
            return (fz + p, gz + 1/s * d)
        else:
            return f(z) + p

    sol = minimize(fg, x, jac=jac)
    for _ in range(4):
        sol = minimize(fg, sol.x, jac=jac)
    return sol


def fd_Jv(f, x, shape=None):
    """Finite-difference approximation to a Jacobian-vector product, J[f](x) v,
    where f: Rⁿ ↦ Rᵐ.

    Example: If f is a gradient, this function approximates Hessian-vector
    products.

    Solve a linear system for second-order gradient methods without
    materializing the Hessian or Fisher matrix. For example, the natural
    gradient direction in an exponential family:

      g = Euclidean_gradient(w)
      Fvp = fd_Jv(dlogZ, w)     # Fisher-vector products
      ng = implicit_cg(Fvp, g)[0]

    """
    if shape is None:
        [N] = x.shape
        [M] = f(x).shape
    else:
        N,M = shape
    def Jv(v, eps=1e-5):
        return (f(x + eps*v) - f(x - eps*v)) / (2*eps)
    return LinearOperator((N, M), matvec=Jv)


def finite_difference(f, eps=1e-5):
    def grad(x):
        x = np.array(x)
        g = []
        for k in range(x.shape[0]):
            v = x[k]
            x[k] = v + eps
            b = f(x)
            x[k] = v - eps
            a = f(x)
            x[k] = v
            g.append((b-a) / 2 / eps)
        return np.array(g)
    return grad


def fdcheck(func, w, g, keys = None, eps = 1e-5, quiet=0, verbose=1, progressbar=1, throw=True):
    """
    Finite-difference check.

    Returns `arsenal.math.compare` instance.

    - `func`: zero argument function, which references `w` in caller's scope.
    - `w`: parameters.
    - `g`: gradient estimate to compare against
    - `keys`: dimensions to check
    - `eps`: perturbation size

    """
    if quiet:
        verbose = 0
        progressbar = 0

    if keys is None:
        if hasattr(w, 'keys'):  # support for sparse vectors represented as a dictionary-like object.
            keys = list(w.keys())
            d = {}
        else:
            # use flat views, if need be.
            if len(w.shape) > 1: w = w.flat
            if len(g.shape) > 1: g = g.flat
            d = np.zeros_like(w)
            keys = list(range(len(w)))    # TODO: these keys have lost their names. So not good for debugging.
    else:
        d = {}

    for k in (iterview(keys) if progressbar else keys):
        was = w[k]
        w[k] = was + eps
        b = func()
        w[k] = was - eps
        a = func()
        w[k] = was
        d[k] = (b-a) / (2*eps)

    if throw and not np.allclose([d[k] for k in keys],
                                 [g[k] for k in keys]):
        compare(d, g, verbose=True)
        raise AssertionError('^^^^ see compare above')


    return compare([d[k] for k in keys],
                   [g[k] for k in keys],
                   verbose=verbose)


def quick_fdcheck(func, w, g, n_checks = 20, eps = 1e-5,
                  quiet=True, verbose=False, progressbar=False, throw=True):
    """
    Check gradient along random directions (a faster alternative to axis-aligned directions).

    Tim Vieira (2017) "How to test gradient implementations"
    https://timvieira.github.io/blog/post/2017/04/21/how-to-test-gradient-implementations/

    """

    if quiet:
        verbose = 0
        progressbar = 0

    keys = ['rand_%s' % i for i in range(n_checks)]
    H = {}
    G = {}

    was = w.flatten()

    w = np.asarray(w.flat)
    g = np.asarray(g.flat)

    dim = len(w)

    for k in (iterview(keys) if progressbar else keys):
        d = spherical(dim)
        G[k] = g.dot(d)
        w[:] = was + eps*d
        b = func()
        w[:] = was - eps*d
        a = func()
        w[:] = was
        H[k] = (b-a) / (2*eps)

    different = not np.allclose(list(H.values()), list(G.values()))
    if verbose or different:
        compare(H, G, verbose=True)

    if different and throw:
        raise AssertionError('^^^^ see compare above')

    return compare(H, G, verbose=False)



from arsenal.maths.util import *
from arsenal.maths.rvs import *
from arsenal.maths.compare import compare, check_equal, pp_plot
from arsenal.maths import stepsize
from arsenal.maths.checkgrad import fdcheck, quick_fdcheck
from arsenal.maths.optimize import NProx, NProj



"""
File system utilities
"""
import re, os, tempfile, shutil
from contextlib import contextmanager
from fnmatch import fnmatch
from arsenal.iterextras import atmost
from subprocess import Popen, PIPE
from path import Path


def filesize(f):
    """
    Uses du to compute human readable summary of filesize.

    It's a wrapper around

    $ du -hs filename

    """
    f = Path(f)
    if not f.exists():
        return 'unknown'
    try:
        return Popen(['du', '-hs', f], stdout=PIPE).communicate()[0].strip().split()[0]
    except TypeError:
        return 'unknown'


def filetype(f):
    try:
        return Popen(['file', '-ib', f], stdout=PIPE).communicate()[0].split(';')[0]
    except TypeError:
        return 'unknown'


def mkdir(d, verbose=False):
    """ Ensure  directories need to create a file exist. """
    try:
        os.makedirs(d)
    except OSError as e:
        if verbose:
            print('[ensuredir]', d, 'suppressing:', e)
        if e.errno != 17:  # errno 17: File exists (you probably don't have permissions)
            raise
    else:
        if verbose:
            print('[ensuredir] created', d)
    return d


def ensure_dir(f, verbose=False):
    """ Ensure  directories need to create a file exist. """
    d = os.path.abspath(os.path.dirname(f))
    mkdir(d, verbose=verbose)
    return d


def clear_dir(d):
    try:
        shutil.rmtree(d)
    except:
        pass
    os.mkdir(d)


@contextmanager
def cd(d=None):
    before = os.getcwd()
    if d is not None:
        os.chdir(d)
    yield
    os.chdir(before)


class preserve_cwd(object):
    """
    context-manager which doubles as a decorator that preserve current
    working directory.

    Usage example:

    As a decorator:
        >>> before = os.getcwd()
        >>> @preserve_cwd
        ... def foo():
        ...     os.chdir('..')
        >>> foo()
        >>> before == os.getcwd()
        True

    As a context-manager:
        >>> before = os.getcwd()
        >>> with preserve_cwd():
        ...     os.chdir('..')
        >>> before == os.getcwd()
        True
    """
    def __init__(self, f=None):
        self.f = f
        self._cwd = None

    def __enter__(self):
        self._cwd = os.getcwd()

    def __exit__(self, *args):
        os.chdir(self._cwd)

    def __call__(self, *args, **kwargs):
        with self:
            return self.f(*args, **kwargs)


@contextmanager
def atomicwrite(filename, mode=0o666, verbose=False):
    """
    Write to `filename` atomically, if for some reason an error occurs in this
    context the contents of the file prior to entering will not be lost.

    Args:
      filename: str; the name of the file
      mode: permissions with which to create the file
    """

    # create the temp file in the same directory
    _, tmp_filename = tempfile.mkstemp(prefix=os.path.basename(filename),
                                       dir=os.path.dirname(filename))

    if verbose:
        print('[atomicwrite] using temporary file:', tmp_filename)

    with open(tmp_filename, 'wb') as f:
        yield f

    try:
        os.chmod(tmp_filename, mode)
        os.rename(tmp_filename, filename)
    except OSError as exc:
        try:
            os.remove(tmp_filename)
        except OSError as e:
            exc = OSError('%s.\n\natomicwrite encountered additional '
                          'errors cleaning up temporary file "%s":\n%s' % (exc, tmp_filename, e))
        raise exc


_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')
_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                         'LPT2', 'LPT3', 'PRN', 'NUL')

def secure_filename(filename):
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.

    On windows system the function also makes sure that the file is not
    named after one of the special device files.

    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'
    >>> secure_filename(u'no brackets [ ] allowed either.txt')
    'no_brackets___allowed_either.txt'

    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you generate random
    filename if the function returned an empty one.
    """
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore').decode('ascii')
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
                   filename.split()))).strip('._')
    filename = re.sub('[\[\]]', '', filename)

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
       filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename


def find_new_title(d, filename):
    """If file *filename* exists in directory `d`, adds or changes the
    end of the file title until a name is found that doesn't yet exist.
    Returns the new file name (without directory).
    For instance, if file "Image (01).jpg" exists,
    returns "Image (02).jpg".
    """
    rx = re.compile(r"\((\d{1,5})\)$")
    p = os.path.join(d, filename)
    while os.path.exists(p):
        base = os.path.basename(p)
        (root, ext) = os.path.splitext(base)
        m = rx.search(root)
        if m == None:
            replacement = "(001)"
        else:
            increment = int(m.group(1)) + 1
            replacement = "(%03d)" % increment
            root = root[:m.start(1)-1]
        f = root + replacement + ext
        p = os.path.join(d, f)
    return p


def files(d, abspath=False):
    "Recursively list all files."
    for dirpath, _, filenames in os.walk(d):
        for f in filenames:
            f = os.path.join(dirpath, f)
            if abspath:
                f = os.path.abspath(f)
            yield f


# TODO: there has to be a faster way to do this which doesn't
# require reading a list of all the files in the directory.
def directories(d, abspath=False):
    """ recursively list all directories. """
    for directory, _, _ in os.walk(d):
        if abspath:
            directory = os.path.abspath(directory)
        yield directory


def find(d, filterfn=None, abspath=False, glob=None, regex=None, dirs=False):
    """
    Recursively walks directory `d` yielding files which satisfy `filterfn`.
    Set option `relpath` to False to output absolute paths.

    glob: shell glob filter function
    regex: regex filter function
    dirs: only search for directories matching filterfn
    """

    assert atmost(1, [filterfn, glob, regex])

    if filterfn is None:
        if glob is not None:
            filterfn = lambda x: fnmatch(x, glob)
        elif regex is not None:
            filterfn = re.compile(regex).match

    collection = directories(d, abspath=abspath) if dirs else files(d, abspath=abspath)

    for item in collection:
        if filterfn is None or filterfn(item):
            yield item


if __name__ == '__main__':

    def run_tests():

        def test_preserve_cwd():
            before = os.getcwd()
            with preserve_cwd():
                os.chdir('..')
                os.chdir('..')
            assert before == os.getcwd()

            @preserve_cwd
            def foo():
                os.chdir('..')
                os.chdir('..')
            cwd_before = os.getcwd()
            foo()
            assert os.getcwd() == cwd_before

        test_preserve_cwd()
        print('done.')

    run_tests()



from scipy.optimize import minimize


class NProx:
    """
    Numerically estimate the proximal operator `Prox_{s*f}(x)` and related
    concepts such as the Moreau envelope.

    This implementation is not efficient and should only be used for testing
    purposes.

    https://web.stanford.edu/~boyd/papers/pdf/prox_algs.pdf
    """

    def __init__(self, f, s):
        self.f = f
        self.s = s

    def solve(self, x, y0=None):
        return minimize(lambda y: self.objective(x, y), y0 if y0 is not None else x)

    def objective(self, x, y):
        d = (x - y)
        return self.f(y) + 0.5/self.s * d @ d

    def op(self, x, **kw):
        return self.solve(x, **kw).x

    def func(self, x, **kw):
        "Proximally smoothed `f` with smoothing parameter `s > 0`."
        return self.solve(x, **kw).fun

    def grad_like(self, x, **kw):
        "Gradient of `prox_func` wrt `x`."
        return (x - self.op(x, **kw)) / self.s


class NProj:

    def __init__(self, C, type):
        self.C = C
        self.type = type
        self.constraints = [{
            'fun': self.C,
            'type': {'>=': 'ineq', '==': 'eq'}[self.type],
        }]

    def solve(self, x, y0=None):
        sol = minimize(
            lambda y: 0.5 * (y - x) @ (y - x),
            y0 if y0 is not None else x,
            constraints = self.constraints,
        )
        if not sol.success: print(sol.message)
        return sol

    def op(self, x, **kw):
        return self.solve(x, **kw).x

    def func(self, x, **kw):
        return self.solve(x, **kw).fun

    def grad_like(self, x, **kw):
        "Gradient of `prox_func` wrt `x`."
        return (x - self.op(x, **kw))


def is_subgradient(f, g, x):
    "Numerically check whether or not `g` is a subgradient of `f` at `x`."
    # Check the condition:
    #    for all y, f(y) - f(x) >= g @ (y - x).
    # iff
    #    for all y, f(y) - g @ y >= f(x) - g @ x
    #         min_y f(y) - g @ y >= f(x) - g @ x
    sol = minimize(lambda y: f(y) - g @ y, x)
    return sol.fun >= f(x) - g @ x



import numpy as np


def mc_perm_test(xs, ys, samples=10000, statistic=np.mean):
    def effect(xs,ys): return np.abs(statistic(xs) - statistic(ys))
    n, k = len(xs), 0.0
    diff = np.abs(np.mean(xs) - np.mean(ys))
    zs = np.concatenate([xs, ys])
    for _ in range(samples):
        np.random.shuffle(zs)
        k += diff <= effect(zs[:n], zs[n:])
    return k / samples


def mc_paired_perm_test(xs, ys, samples=10000, statistic=np.mean):
    """
    Paired permutation test

      >>> xs = np.array([1,2,3,4,5,6])
      >>> ys = np.array([2,3,4,5,6,7])
      >>> u = mc_perm_test(xs, ys, 1000)
      >>> p = mc_paired_perm_test(xs, ys, 1000)

    Under the unpaired test, we do not have a significant difference because the
    systems only differ in two positions.

      >>> assert u > .40

    Under the Paired test, we have more power!
      >>> assert p < .05

    """
    def effect(xs,ys): return np.abs(statistic(xs) - statistic(ys))
    assert len(xs) == len(ys)
    n, k = len(xs), 0
    diff = effect(xs,ys)    # observed difference
    for _ in range(samples):  # for each random sample
        swaps = np.random.randint(0,2,n).astype(bool)     # flip n coins
        k += diff <= effect(np.select([swaps,~swaps],[xs,ys]), # swap elements accordingly
                            np.select([~swaps,swaps],[xs,ys]))
    return k / samples  # fraction of random samples that achieved at least the observed difference


from itertools import product
def bf_paired_perm_test(xs, ys, statistic=np.mean):
    def effect(xs,ys): return np.abs(statistic(xs) - statistic(ys))
    assert len(xs) == len(ys)
    observed = effect(xs, ys)
    p = 0.0; n = len(xs)
    for swaps in product(*([0,1] for _ in range(n))):
        swaps = np.array(swaps, dtype=bool)
        pe = 2**-n
        E = effect(np.select([swaps,~swaps],[xs,ys]), # swap elements accordingly
                   np.select([~swaps,swaps],[xs,ys]))
        p += pe * (E >= observed)
    return p


def verbose_paired_perm_test(xs, ys, nmc=10_000, threshold=0.05, fmt='%.4f'):
    "Let xs be the system you want be greater."
    from arsenal import colors

    p = mc_paired_perm_test(xs, ys, nmc)

    mx = np.mean(xs)
    my = np.mean(ys)
    if p <= threshold:
        if mx > my:
            c = colors.green
            d = '>'
        else:
            c = colors.red
            d = '<'
    else:
        c = colors.yellow
        d = '~'

    #print('brute-force', bf_paired_perm_test(xs, ys))

    print('[paired perm] %s (p=%s)' % (c % 'X (%s) %s Y (%s)' % (fmt % mx,
                                                                 d,
                                                                 fmt % my),
                                       fmt % p))

    return p


if __name__ == '__main__':
    def test():
        xs = np.array([1,2,3,4,5,6])
        verbose_paired_perm_test(xs+1, xs)
        verbose_paired_perm_test(xs, xs+1)
        verbose_paired_perm_test(xs, xs)
    test()



"""
OrderedSet -- a set which remembers insertion order.
"""

class OrderedSet:
    """
    Set which remembers insertion ordering allowed iteration while changing size
    and determinism.
    """
    __slots__ = 'set', 'list'
    def __init__(self, elems=None):
        self.set = set()
        self.list = []
        if elems is not None:
            for x in elems:
                self.add(x)
    def __contains__(self, item):
        return item in self.set
    def __iter__(self):
        return iter(self.list)
    def add(self, item):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)
    def __len__(self):
        return len(self.set)
    def __repr__(self):
        return 'OrderedSet(%r)' % self.list
    def __getitem__(self, *args):
        return self.list.__getitem__(*args)
    def __sub__(self, other):
        new = OrderedSet()
        for x in self:
            if x not in other:
                new.add(x)
        return new
    def __or__(self, other):
        new = OrderedSet()
        for x in self:
            new.add(x)
        for x in other:
            new.add(x)
        return new
    def __ior__(self, other):
        for x in other:
            self.add(x)
        return self
    def isdisjoint(self, other):
        return self.set.isdisjoint(other.set)
    def __lt__(self, other):
        return self.set < other.set
    def __le__(self, other):
        return self.set <= other.set
    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return self.set == other.set
        elif isinstance(other, set):
            return self.set == other
        else:
            return False
    def remove(self, x):
        "this method is slow; avoid"
        self.set.remove(x)
        self.list = [y for y in self.list if x != y]



class AbstractIntegerizer:

    def __getitem__(self, i):
        if isinstance(i, list):
            return [self._list[ii] for ii in i]
        return self._list[i]

    def __call__(self, k):
        if isinstance(k, list):
            ks = k
            return [self._add(k) for k in ks]
        return self._add(k)

    def _add(self, k):
        raise NotImplementedError()

    # method aliases
    encode = __call__
    decode = __getitem__
    lookup = __getitem__
    add = __call__


class Integerizer(AbstractIntegerizer):
    """
    Class for maintaining a perfect hash for a set of keys.

    >>> a = Integerizer()
    >>> a(list('abcd'))
    [0, 1, 2, 3]

    >>> a(list('ace'))
    [0, 2, 4]

    >>> a[list(range(4))]
    ['a', 'b', 'c', 'd']

    >>> a.freeze()
    >>> a.add('z')
    Traceback (most recent call last):
      ...
    ValueError: Alphabet is frozen. Key "z" not found.

    >>> list(a)
    ['a', 'b', 'c', 'd', 'e']

    """

    def __init__(self, data=None):
        self._map = {}       # str -> int
        self._list = []      # int -> str
        self._frozen = False
        if data: self.add(data)

    def _add(self, k):
        try:
            return self._map[k]
        except KeyError:
            #if not isinstance(k, basestring):
            #    raise ValueError("Invalid key (%s): only strings allowed." % (k,))
            if self._frozen:
                raise ValueError('Alphabet is frozen. Key "%s" not found.' % (k,))
            x = self._map[k] = len(self._list)
            self._list.append(k)
            return x

    def __contains__(self, k):
        return k in self._map

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        return self._list == other._list

    def __repr__(self):
        return 'Alphabet(size=%s,frozen=%s)' % (len(self), self._frozen)

    def freeze(self):
        self._frozen = True
        return self

    def keys(self):
        return self._list

    def items(self):
        return self._map.items()


def jenkins32(a):
    assert isinstance(a, int) and a >= 0
    a = (a+0x7ed55d16) + (a<<12)
    a = (a^0xc761c23c) ^ (a>>19)
    a = (a+0x165667b1) + (a<<5)
    a = (a+0xd3a2646c) ^ (a<<9)
    a = (a+0xfd7046c5) + (a<<3)
    a = (a^0xb55a4f09) ^ (a>>16)
    return a


# TODO: might be good to write in Cython and provide a few fast hash functions
# for strings, ints, and tuples.
#   ^ I have a reasonable implementation in dyna.learn.features
class FeatureHashing(AbstractIntegerizer):
    """
    >>> h = FeatureHashing(jenkins32, 8)
    >>> h([1, 2, 3])
    [182, 76, 156]

    """

    def __init__(self, h, bits):
        self.h = h
        self.bits = bits
        self.D = 2**bits
        self.mask = self.D - 1   # use bit mask rather than mod for efficiency

    def _add(self, k):
        h = self.h(k)
        assert h >= 0
        return h & self.mask


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import numpy as np
import matplotlib.pyplot as pl
import scipy.stats as st
from numpy import array, exp, cumsum, asarray
from numpy.linalg import norm
from scipy.integrate import quad


def is_distribution(p):
    p = asarray(p)
    return (p >= 0).all() and abs(1 - p.sum()) < 1e-10


def anneal(p, *, invT=None, T=None):
    "p ** (1/T)"
    if T is not None:
        assert invT is None
        invT = 1./T
    p = p ** invT
    p /= p.sum()
    return p


class Max:
    def __init__(self, X, Y):
        self.X = X
        self.Y = Y

    def cdf(self, t):
        return self.X.cdf(t)*self.Y.cdf(t)

    def pdf(self, t):
        X = self.X; Y = self.Y
        #  d/dt[cdf(t)] = d/dt[X.cdf(t)*Y.cdf(t)]
        #               = X.cdf(t)*d/dt[Y.cdf(t)] + d/dt[X.cdf(t)]*Y.cdf(t)
        #               = X.cdf(t)*Y.pdf(t) + X.pdf(t)*Y.cdf(t)
        return X.pdf(t)*Y.cdf(t) + X.cdf(t)*Y.pdf(t)

    def ppf(self, u):
        raise ValueError('no closed form for ppf(max(X,Y))')

    def rvs(self, size):
        return np.maximum(self.X.rvs(size), self.Y.rvs(size))


class _BruteForce:
    """
    Base class for tabular representation of a distribution

      p(x) ∝ score(x).

    where score(x) > 0 for all x ∈ domain.

    """

    def __init__(self):
        self.Z = np.sum([self.score(x) for x in self.domain()])
        self.P = {x: self.score(x) / self.Z for x in self.domain()}

    def domain(self):
        raise NotImplementedError

    def score(self, x):
        raise NotImplementedError()

    def entropy(self):
        return -np.sum([p * np.log(p) for p in self.P.values() if p != 0])

    def logp(self, x):
        return np.log(self.P[x])


class TruncatedDistribution:
    def __init__(self, d, a, b):
        assert np.all(a <= b), [a, b]
        self.d = d; self.a = a; self.b = b
        self.cdf_b = d.cdf(b)
        self.cdf_a = d.cdf(a)
        self.cdf_w = self.cdf_b - self.cdf_a
    def sf(self, x):
        return 1-self.cdf(x)
    def pdf(self, x):
        return (self.a < x) * (x <= self.b) * self.d.pdf(x) / self.cdf_w
    def rvs(self, size=None):
        u = np.random.uniform(0, 1, size=size)
        return self.ppf(u)
    def ppf(self, u):
        return self.d.ppf(self.cdf_a + u * self.cdf_w)
    def cdf(self, x):
        return np.minimum(1, (self.a < x) * (self.d.cdf(x) - self.cdf_a) / self.cdf_w)
    def mean(self):
        # The truncated mean is unfortunately not analytical
        return quad(lambda x: x * self.pdf(x), self.a, self.b)[0]

        # XXX: The Darth Vader only applies to positive random variables
        # http://thirdorderscientist.org/homoclinic-orbit/2013/6/25/the-darth-vader-rule-mdash-or-computing-expectations-using-survival-functions
        # https://content.sciendo.com/view/journals/tmmp/52/1/article-p53.xml
#        return quad(self.sf, self.a, self.b)[0] + self.a



# References
#
#  - Hypoexponential
#    https://en.wikipedia.org/wiki/Hypoexponential_distribution
#
#  - Márton Balázs, https://people.maths.bris.ac.uk/~mb13434/sumexp.pdf
#
#  - Bibinger (2013) "Notes on the sum and maximum of independent exponentially
#    distributed random variables with different scale parameters"
#
#  - https://actuarialmodelingtopics.wordpress.com/2016/08/01/the-hyperexponential-and-hypoexponential-distributions/
#
class SumOfExponentials:
    def __init__(self, w):
        self.w = w
        [self.n] = w.shape
        self.a = np.prod(w)
        self.B = np.array([np.prod([w[j] - w[i] for j in range(self.n) if j != i]) for i in range(self.n)])
    def cdf(self, x):
        return 1-self.a*sum(np.exp(-self.w[j]*x) / (self.w[j] * self.B[j]) for j in range(self.n))
    def pdf(self, x):
        return self.a * sum(np.exp(-self.w[j]*x) / self.B[j] for j in range(self.n))
    def rvs(self, size=None):
        if size is not None: size = (size, self.n)
        return -np.sum(np.log(np.random.uniform(0,1,size=size)) / self.w[None,:], axis=1)
    def mgf(self, t):
        return self.a / np.prod(self.w - t)
    def d_mgf(self, t, n):
        # TODO: work out the derivative
        return nd.Derivative(self.mgf, n=n)(t)
    def mean(self):
        return np.sum(1/self.w) # sum of the means
    def var(self):
        return np.sum(1/self.w**2) # sum of the variances because they are indep


def show_distr(D, a, b, resolution=1000):
    xs = np.linspace(a, b, resolution)
    us = np.linspace(0, 1, resolution)
    fig, ax = pl.subplots(figsize=(12,4), ncols=3)

    if hasattr(D, 'pdf'):
        ax[0].plot(xs, D.pdf(xs))
        ax[0].set_ylabel('f(x)')
        ax[0].set_xlabel('x')
        ax[0].set_title('probability density function')
        ax[0].set_xlim(a, b)

    ax[1].plot(xs, D.cdf(xs))
    ax[1].set_ylabel('F(x)')
    ax[1].set_xlabel('x')
    ax[1].set_title('cumulative distribution function')
    ax[1].set_xlim(a, b)


    if hasattr(D, 'ppf'):
        ax[2].plot(us, D.ppf(us))
        ax[2].set_ylabel('$F^{-1}(u)$')
        ax[2].set_xlabel('u')
        ax[2].set_xlim(0, 1)
        ax[2].set_title('quantile function')

    fig.tight_layout()

    for a in ax:
        # Move left and bottom spines outward by 10 points
        a.spines['left'].set_position(('outward', 10))
        a.spines['bottom'].set_position(('outward', 10))
        # Hide the right and top spines
        a.spines['right'].set_visible(False)
        a.spines['top'].set_visible(False)
        # Only show ticks on the left and bottom spines
        a.yaxis.set_ticks_position('left')
        a.xaxis.set_ticks_position('bottom')

    return ax


def compare_samples_to_distr(D, samples, a, b, bins):

    fig, ax = pl.subplots(figsize=(12,4), ncols=3)

    ax[0].hist(samples, bins=bins, color='b', alpha=0.5, density=True,
               label='histogram')

    xs = np.linspace(a, b, 1000)
    ax[0].plot(xs, D.pdf(xs), c='k')
    ax[0].set_title('pdf')

    E = Empirical(samples)

    ax[1].plot(xs, E.cdf(xs), alpha=0.5, linestyle=':')
    ax[1].plot(xs, D.cdf(xs), alpha=0.5)
    ax[1].set_title('cdf')

    us = np.linspace(0, 1, 1000)
    ax[2].plot(us, E.ppf(us), alpha=0.5, linestyle=':')
    ax[2].plot(us, [D.ppf(u) for u in us], alpha=0.5)
    ax[2].set_title('ppf')

    for a in ax:
        # Move left and bottom spines outward by 10 points
        a.spines['left'].set_position(('outward', 10))
        a.spines['bottom'].set_position(('outward', 10))
        # Hide the right and top spines
        a.spines['right'].set_visible(False)
        a.spines['top'].set_visible(False)
        # Only show ticks on the left and bottom spines
        a.yaxis.set_ticks_position('left')
        a.xaxis.set_ticks_position('bottom')

    fig.tight_layout()

    return ax


def test_truncated_distribution():
    import matplotlib.pyplot as pl
    import scipy.stats as st
    d = st.lognorm(1.25)

    t = TruncatedDistribution(d, 2, 4)

    print(t.mean(), t.rvs(100_000).mean())

    if 1:
        us = np.linspace(0.01, 0.99, 1000)
        xs = np.linspace(d.ppf(0.01), d.ppf(0.99), 1000)
        pl.plot(xs, t.cdf(xs), label='cdf')
        pl.plot(t.ppf(us), us, label='ppf')
        pl.legend(loc='best')
        pl.xlim(0, 6)
        pl.show()

    if 1:
        pl.hist(t.rvs(100000), density=True, bins=200)
        pl.plot(xs, t.pdf(xs), c='r', lw=3, alpha=0.75)
        pl.xlim(0, 6)
        pl.show()

    if 1:
        us = np.linspace(0.01, 0.99, 100)
        for u in us:
            v = t.cdf(t.ppf(u))
            assert abs(u - v)/abs(u) < 1e-3

    if 1:
        xs = np.linspace(1, 5, 1000)
        for x in xs:
            y = t.ppf(t.cdf(x))
            if t.pdf(x) > 0:
                err = abs(x - y)/abs(x)
                assert err < 1e-3, [err, x, y]


def random_dist(*size):
    """
    Generate a random conditional distribution which sums to one over the last
    dimension of the input dimensions.
    """
    return np.random.dirichlet(np.ones(size[-1]), size=size[:-1])


def random_psd(n):
    return st.wishart.rvs(df=n, scale=np.eye(n)) / n


class Mixture(object):
    """
    Mixture of several densities
    """
    def __init__(self, w, ds):
        w = array(w)
        assert is_distribution(w), \
            'w is not a prob. distribution.'
        self.ds = ds
        self.w = w

    def rvs(self, size=1):
        # sample component
        i = sample(self.w, size=size)
        # sample from component
        return array([self.ds[j].rvs() for j in i])

    def pdf(self, x):
        return sum([d.pdf(x) * w for w, d in zip(self.w, self.ds)])

    def ppf(self, q):
        from scipy.optimize import minimize
        x0 = sum([d.ppf(q) * w for w, d in zip(self.w, self.ds)])
        return minimize(lambda t: np.sum((self.cdf(t) - q)**2), x0).x

    def cdf(self, x):
        return sum([d.cdf(x) * w for w, d in zip(self.w, self.ds)])

    def mean(self):
        return sum([d.mean() * w for w, d in zip(self.w, self.ds)])


def spherical(size):
    "Generate random vector from spherical Gaussian."
    x = np.random.normal(0, 1, size=size)
    x /= norm(x, 2)
    return x



# TODO: Should we create a class to represent this data with this the fit
# method?  This should really be the MLE of some sort of distrbution.  What
# distribution is it?  I suppose it is nonparametric in the same sense that
# Kaplan-Meier is nonparametric (In fact, KM generalizes this estimator to
# support censored response).
# TODO: That same class would probably have the defacto mean/std estimators.
class Empirical:
    """
    Empirical CDF of data `a`, returns function which makes values to their
    cumulative probabilities.

     >>> g = cdf([5, 10, 15])

    Evaluate the CDF at a few points

     >>> g([5,9,13,15,100])
     array([0.33333333, 0.33333333, 0.66666667, 1.        , 1.        ])


    Check that ties are handled correctly

     >>> g = cdf([5, 5, 15])

     The value p(x <= 5) = 2/3

     >>> g([0, 5, 15])
     array([0.        , 0.66666667, 1.        ])

    The auantile function should be the inverse of the cdf.

      >>> g = cdf([-1, 5, 5, 15])
      >>> g.quantile(np.linspace(0, 1, 10))
      array([-1, -1, -1,  5,  5,  5,  5,  5,  5, 15])

    """

    def __init__(self, x):
        self.x = x = np.array(x, copy=True)
        [self.n] = x.shape
        self.x.sort()

    def __call__(self, z):
        return self.x.searchsorted(z, 'right') / self.n

    cdf = __call__

    def sf(self, z):
        return 1-self.cdf(z)

    def conditional_mean(self, a, b):
        "E[T | a <= T < b]"
        m = 0.0; n = 0.0
        for i in range(self.x.searchsorted(a, 'right'),
                       self.x.searchsorted(b, 'left')):
            m += self.x[i]
            n += 1
        return m / n if n > 0 else np.inf

    def quantile(self, q):
        # TODO: this could be made fastet given that x is already sorted.
        assert np.all((0 <= q) & (q <= 1))
        return np.quantile(self.x, q, interpolation='lower')

    def plot(self, confidence=0.95, color=None, label=None):
        pl.plot(self.x, self.cdf(self.x), label=label, c=color)

        if confidence is not None:
            # show confidence bands (derived from the DKW inequality)
            # https://bjlkeng.github.io/posts/the-empirical-distribution-function/
            # https://en.wikipedia.org/wiki/Dvoretzky%E2%80%93Kiefer%E2%80%93Wolfowitz_inequality
            eps = np.sqrt(np.log(2 / (1-confidence)) / (2 * self.n))
            pl.fill_between(self.x,
                            np.maximum(0, self(self.x) - eps),   # lower
                            np.minimum(1, self(self.x) + eps),   # upper
                            alpha=0.1,
                            color=color,
                            label=f'{label} {confidence:.2f} confidence band')

    ppf = quantile

cdf = Empirical


def sample(w, size=None, u=None):
    """
    Uses the inverse CDF method to return samples drawn from an (unnormalized)
    discrete distribution.
    """
    c = cumsum(w)
    if u is None:
        u = np.random.uniform(0,1,size=size)
    return c.searchsorted(u * c[-1])


def log_sample(w, **kwargs):
    "Sample from unnormalized log-distribution."
    a = np.array(w, dtype=float)
    a -= a.max()
    exp(a, out=a)
    return sample(a, **kwargs)


def sample_dict(x, *args, **kwargs):
    "Sample keys from `x` in proportion to their values."
    keys = np.array(list(x.keys()), dtype=object)
    vals = np.array(list(x.values()))
    return keys[sample(vals, *args, **kwargs)]


def test_mixture():

    d = [st.norm(1, 2), st.norm(2, .5), st.norm(3, .1)]

    w = np.array([.2, .7, .1])

    D = Mixture(w, d)

    S = D.rvs(100_000)

    assert abs(D.mean() - S.mean()) < 0.005, abs(D.mean() - S.mean())

    a, b = -5, 5

    compare_samples_to_distr(D, S, a, b, bins=100)
    pl.show()

#    if 1:
#        us = np.linspace(0.01, 0.99, 100)
#        for u in us:
#            v = D.cdf(D.ppf(u))
#            err = abs(u - v)/abs(u)
#            assert err < 1e-3, err
#
#    if 1:
#        xs = np.linspace(a, b, 100)
#        for x in xs:
#            y = D.ppf(D.cdf(x))
#            if D.pdf(x) > 0:
#                err = abs(x - y)/abs(x)
#                assert err < 1e-3, [err, x, y]


def test_cdf():
    n = 1_000
    x = np.random.normal(0,1,size=n)
    cdf(x).plot()
    pl.legend(loc='best')
    pl.show()


def test_sample_dict():

    ws = {'a': 9, 'b': 1, 'c': 0}
    W = sum(ws.values())

    xs = sample_dict(ws, size=10)
    assert len(xs) == 10

    N = 1_000_000
    xs = sample_dict(ws, size=N)
    from collections import Counter
    c = Counter(xs)
    for x in ws:
        print(ws[x]/W, c[x]/N)
        assert abs(ws[x]/W - c[x]/N) <= 0.001

    x = sample_dict(ws, u=.5)
    assert x == 'a'

    x = sample_dict(ws, u=.91)
    assert x == 'b'


if __name__ == '__main__':
    #test_truncated_distribution()
    #test_mixture()
    from arsenal import testing_framework
    testing_framework(globals())



import numpy as np
import matplotlib.pyplot as pl

from arsenal.maths import compare, random_dist
from collections import Counter
from arsenal.maths.combinatorics import permute
from arsenal.timer import timers

# XXX: warning some of these implementations may not be shipped with the
# package.  Also be sure to compile with Cython.
from arsenal.datastructures.heap.sumheap  import SumHeap as SumHeap1
from arsenal.datastructures.heap.sumheap2 import SumHeap as SumHeap2
from arsenal.datastructures.heap.sumheap3 import SumHeap as SumHeap3


def p_perm(w, z):
    "The probability of a permutation `z` under the sampling without replacement scheme."
    n = len(w); k = len(z)
    assert 0 < k <= n
    wz = w[np.array(z, dtype=int)]
    W = wz[::-1].cumsum()
    return np.prod(wz / W)


def swor_heap1(w, R):
    n = len(w)
    z = np.zeros((R, n), dtype=int)
    for r in range(R):
        h = SumHeap1(w)
        z[r] = h.swor(n)
    return z


def swor_heap2(w, R):
    n = len(w)
    z = np.zeros((R, n), dtype=int)
    for r in range(R):
        h = SumHeap2(w)
        z[r] = h.swor(n)
    return z


def swor_heap3(w, R):
    n = len(w)
    z = np.zeros((R, n), dtype=int)
    for r in range(R):
        h = SumHeap3(w)
        z[r] = h.swor(n)
    return z


def counts(S):
    "empirical distribution over z"
    c = Counter()
    m = len(S)
    for s in S:
        c[tuple(s)] += 1 / m
    return c


def test():
    methods = [
        swor_heap1,
#        swor_heap2,
        swor_heap3,
    ]

    R = 50_000
    v = random_dist(4)

    S = {f.__name__: f(v, R) for f in methods}

    D = {name: counts(S[name]) for name in S}

    R = {}
    n = len(v)
    for z in permute(range(n)):
        R[z] = p_perm(v, z)
        for d in D.values():
            d[z] += 0

    # Check that p_perm sums to one.
    np.testing.assert_allclose(sum(R.values()), 1)
    for name, d in sorted(D.items()):
        compare(R, d)#.show(title=name);

    T = timers()
    R = 50
    for i in range(1, 15):
        n = 2**i
        #print('n=', n, 'i=', i)
        for _ in range(R):
            v = random_dist(n)
            np.random.shuffle(methods)
            for f in methods:
                name = f.__name__
                with T[name](n = n):
                    S = f(v, R = 1)
                assert S.shape == (1, n)   # some sort of sanity check
    print('done')

    fig, ax = pl.subplots(ncols=2, figsize=(12, 5))
    T.plot_feature('n', ax=ax[0])
    fig.tight_layout()
    T.plot_feature('n', ax=ax[1]); ax[1].set_yscale('log'); ax[1].set_xscale('log')
    T.compare()

    pl.show()

if __name__ == '__main__':
    test()




from psutil import Process
from os import getpid


def memory_usage():
    "Return the memory usage of this process in MB."
    p = Process(getpid())
    return p.memory_info()[0] / 2**20


if __name__ == '__main__':
    print('memory usage: %.2f MB' % memory_usage())



import signal
import sys
import traceback
from arsenal import colors
from arsenal.timer import htime, time

from argparse import ArgumentParser
from contextlib import contextmanager
from IPython.core import ultratb

from arsenal.maths import restore_random_state


class TestingFramework:

    def __init__(self, tests):

        p = ArgumentParser()
        p.add_argument('filters', nargs='*',
                       help='Include tests which contains any of these substrings.')
        p.add_argument('--skip', nargs='*',
                       help='Exclude tests which contains any of these substrings.',
                       default=[])

        # Random seed control
        p.add_argument('--seed', type=int, default=None)
        p.add_argument('--seed-scan', type=int, default=None)
        self.p = p

        self.overview = OverviewManager(True, False)
        self.tests = tests
        self.cli = p

    @property
    def args(self):
        return self.cli.parse_args()

    def run(self):
        # figure out which random seeds the user wants
        args = self.args
        if args.seed_scan is not None:
            # random seed search for trying to find a failing example.
            assert args.seed_scan > 0
            start_seed = args.seed or 0
            seeds = range(start_seed, start_seed + args.seed_scan)
        else:
            seeds = [args.seed]

        for name in self.find_tests():
            print(colors.light.yellow % colors.thick_line(80))
            print(colors.light.yellow % f'{name}')
            for seed in seeds:
                with restore_random_state(seed):
                    with self.overview(f'{name} (seed={seed})' if seed is not None else f'{name}'):
                        self.tests[name]()
            print()
        self.overview.report()

    def find_tests(self):
        args = self.args
        filters = list(args.filters) if args.filters else ['test_', 'todo_']
        for kw in filters:
            kw = kw.lower()
            for name in sorted(self.tests):
                x = name.lower()
                if (kw in x
                    and (x.startswith('test_') or x.startswith('todo_'))
                    and not any((z in x) for z in args.skip)
                ):
                    yield name


def testing_framework(tests):
    return TestingFramework(tests).run()


class FailedTest(Exception): pass


class OverviewManager:

    def __init__(self, overview, fancy_traceback=False):
        self.passed = 0
        self.failed = 0
        self.overview = overview
        self.fancy_traceback = fancy_traceback
        self.summary = []
        self.callback = None

        # Hardcoded log file
        #self.f = open('failed', 'w')   # track recent failure tests and run those first.

        # When the user send sigquit, we will still try to print the test
        # summary (i.e., call `report` with whatever progress was made).
        #
        # TODO: it'd be better if we also printed which tests did not get run,
        # but the current scheme, but the current usage does not make that easy.
        signal.signal(signal.SIGQUIT, self.SIGQUIT)

        self.suppress_exceptions = (KeyboardInterrupt,)

    def SIGQUIT(self, sig, frame):   # pylint: disable=unused-argument
        self.report()
        exit()

    @contextmanager
    def __call__(self, name):
        print()
        print(colors.yellow % colors.line(80))
        print(colors.yellow % 'Running %s' % name)

        b4 = time()
        if self.overview:
            try:
                yield
            except (KeyboardInterrupt, Exception) as e:          # pylint: disable=broad-except
                if self.callback is not None: self.callback(e)
                if self.fancy_traceback:
                    #exc = ultratb.FormattedTB(mode='Context')
                    exc = ultratb.FormattedTB(mode='Verbose')
                    exc(*sys.exc_info())
                else:
                    msg = ''.join(traceback.format_exception(*sys.exc_info()))
                    if isinstance(e, self.suppress_exceptions):
                        print(e.__class__.__name__, msg)
                    elif isinstance(e, (AssertionError, FailedTest)):
                        print(colors.render(colors.light.red % msg))
                    else:
                        print(colors.render(colors.red % msg))
                self.summary.append('%s %s: %s %s' % (colors.red % 'fail', name, type(e).__name__, colors.render(e)))
                self.failed += 1
                #print(name, file=self.f); self.f.flush()

            else:
                self.summary.append('%s %s' % (colors.green % 'pass', name))
                self.passed += 1
        else:
            yield

        took = time() - b4
        t = colors.magenta % f'({took:.3g}s)'
        print(colors.yellow % f'done {t}')

    def report(self):
        if not self.overview: return
        print()
        print(colors.light.yellow % 'Summary')
        print(colors.light.yellow % colors.thick_line(80))
        for x in sorted(self.summary, key=lambda x: 'fail' in x):
            print(x)
        print()
        print(f'passed: {self.passed}, failed: {self.failed}.')


if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    sys.argv = sys.argv[1:]
    with open(filename) as f: content = f.read()
    global_vars = {}
    exec(content, global_vars)
    testing_framework(global_vars)



"""
Adaptive stepsize algorithms
"""
import numpy as np


__author__ = 'Tim Vieira (http://timvieira.github.io)'


def norm_clip(x, C):
    "Rescale x (in place) so that ||x|| <= C."
    z = np.linalg.norm(x)
    if z > C:
        x[:] *= C/z


def ewma(x, y, alpha):
    "Exponentially weighted moving average (`x` is updated in place)."
    x[:] *= (1-alpha)
    x[:] += alpha*y


class adagrad(object):
    """
    Adagrad
    """
    def __init__(self, x, damping = 1e-4):
        self.x = x
        self.G = np.zeros_like(x) + damping
        self.i = 0
        self.D = None

    def __call__(self, g, learning_rate=1.0):
        # Update parameters
        self.G += g*g
        self.D = learning_rate/np.sqrt(self.G)
        self.x[:] -= g * self.D
        self.i += 1


class sgd(object):
    """
    Simple sgd
    """
    def __init__(self, x):
        self.x = x
        self.i = 0

    def __call__(self, g, learning_rate=1.0):
        self.x[:] -= g * learning_rate
        self.i += 1


class adam(object):
    """
    Adam as described in http://arxiv.org/pdf/1412.6980.pdf.
    """
    def __init__(self, x):
        self.x = x
        self.m = np.zeros_like(x)
        self.v = np.zeros_like(x)
        self.i = 1
        self.D = None

    def __call__(self, g, learning_rate=0.01, b1 = 0.1, b2 = 0.01, lam=1e-4, damping=1e-4):
        i = self.i
        b1t = 1 - (1-b1)*(lam**i)          # b1t -> 1 over time
        # Update first moment estimate
        ewma(self.m, g, b1t)
        # Update second moment estimate
        ewma(self.v, g*g, b2)
        # Bias correction.
        mhat = self.m / (1 - (1-b1)**i)
        vhat = self.v / (1 - (1-b2)**i)
        # Take step
        self.x[:] -= learning_rate*mhat/(np.sqrt(vhat) + damping)
        self.i += 1


class sgd_momentum(object):

    def __init__(self, x):
        self.velocity = np.zeros(len(x))
        self.x = x

    def __call__(self, g, mass=0.9, learning_rate=0.1):
        self.velocity[:] = mass * self.velocity - (1.0 - mass) * g
        self.x[:] += learning_rate * self.velocity
